#include <icd/support/icd_log.h>
#include <icd/icd_gconf.h>

#include <string.h>

#include "gprs-private.h"
#include "gprs.h"
#include "context.h"
#include "observer.h"

typedef struct
{
  gprs_private *priv;
  gchar *key;
  gchar *val;
} property_item_2;

void
close_all_contexts(gprs_private *priv, const gchar *err_str,
                   const gchar *err_note)
{
  GSList *l;

  for (l = priv->networks; l; l = l->next)
  {
    gprs_network *gprs = l->data;

    if (!gprs_network_is_always_online(gprs) && gprs->field_14 <= 2)
      send_ui_request_without_reply(err_note);

    priv->close_fn(ICD_NW_ERROR, err_str, gprs->network_type,
                   gprs->network_attrs, gprs->network_id);
  }
}

gboolean
send_ui_request_without_reply(const char *err_note)
{
  DBusMessage *mcall;
  gboolean rv = FALSE;

  if (err_note)
  {
    mcall = dbus_message_new_method_call("com.nokia.cellular_ui",
                                         "/com/nokia/cellular_ui",
                                         "com.nokia.cellular_ui",
                                         "show_error_note");
    if (mcall)
    {
      if (dbus_message_append_args(mcall,
                                   DBUS_TYPE_STRING, &err_note,
                                   DBUS_TYPE_INVALID))
      {
        icd_dbus_send_system_mcall(mcall, -1, NULL, NULL);
        rv = TRUE;
      }
      else
      {
        ILOG_ERR(
              "[GPRS] could not append args to 'show_error_note' method call");
      }

      dbus_message_unref(mcall);
    }
    else
      ILOG_ERR("[GPRS] could not create 'show_error_note' method call");
  }

  return rv;
}

static gboolean
network_process_all_properties(DBusMessageIter *iter, gprs_network *gprs)
{
  DBusMessageIter sub3;
  DBusMessageIter sub2;
  DBusMessageIter sub1;
  dbus_message_iter_recurse(iter, &sub1);

  while (dbus_message_iter_get_arg_type(&sub1) == DBUS_TYPE_DICT_ENTRY)
  {
    const char *key = NULL;
    int type;

    dbus_message_iter_recurse(&sub1, &sub2);

    if (dbus_message_iter_get_arg_type(&sub2) != DBUS_TYPE_STRING)
      return FALSE;

    dbus_message_iter_get_basic(&sub2, &key);
    dbus_message_iter_next(&sub2);

    if (dbus_message_iter_get_arg_type(&sub2) != DBUS_TYPE_VARIANT)
      return FALSE;

    dbus_message_iter_recurse(&sub2, &sub3);
    type = dbus_message_iter_get_arg_type(&sub3);

    if (!strcmp(key, "APN") || !strcmp(key, "Username") ||
        !strcmp(key, "Password") || !strcmp(key, "PDPType") ||
        !strcmp(key, "Connected"))
    {
      ILOG_INFO("[GPRS] Read key '%s'", key);
    }
    else
    {
      if (!strcmp(key, "PDPAddress"))
      {
        const char *s = NULL;

        if (type != DBUS_TYPE_STRING)
          return FALSE;

        dbus_message_iter_get_basic(&sub3, &s);
        ILOG_INFO("[GPRS] Read key '%s' and value '%s'", key, s);
        g_free(gprs->ip_address);
        gprs->ip_address = g_strdup(s);
      }
      else if (!strcmp(key, "PDNSAddress"))
      {
        const char *s = NULL;

        if (type != DBUS_TYPE_STRING)
          return FALSE;

        dbus_message_iter_get_basic(&sub3, &s);
        ILOG_INFO("[GPRS] Read key '%s' and value '%s'", key, s);
        g_free(gprs->ip_dns1);
        gprs->ip_dns1 = g_strdup(s);
      }
      else if (!strcmp(key, "SDNSAddress"))
      {
        const char *s = NULL;

        if (type != DBUS_TYPE_STRING)
          return FALSE;

        dbus_message_iter_get_basic(&sub3, &s);
        ILOG_INFO("[GPRS] Read key '%s' and value '%s'", key, s);
        g_free(gprs->ip_dns2);
        gprs->ip_dns2 = g_strdup(s);
      }
      else if (!strcmp(key, "NetIF"))
      {
        const char *s = NULL;

        if (type != DBUS_TYPE_STRING)
          return FALSE;

        dbus_message_iter_get_basic(&sub3, &s);

        ILOG_INFO("[GPRS] Read key '%s' and value '%s'", key, s);
        g_free(gprs->gprs_iface);
        gprs->gprs_iface = g_strdup(s);
      }
      else if (!strcmp(key, "RxBytes"))
      {
        dbus_uint64_t bytes;

        if (type != DBUS_TYPE_UINT64)
          return FALSE;

        dbus_message_iter_get_basic(&sub3, &bytes);
        ILOG_INFO("[GPRS] Read key '%s' and value '%llu'", key,
                  (unsigned long long)bytes);
        gprs->rx_bytes = bytes;
      }
      else if (!strcmp(key, "TxBytes"))
      {
        dbus_uint64_t bytes;

        if (type != DBUS_TYPE_UINT64)
          return FALSE;

        dbus_message_iter_get_basic(&sub3, &bytes);
        ILOG_INFO("[GPRS] Read key '%s' and value '%llu'", key,
                  (unsigned long long)bytes);
        gprs->tx_bytes = bytes;
      }
      else
        ILOG_ERR("[GPRS] Error: Unknown key '%s'", key);
    }

    dbus_message_iter_next(&sub1);
  }

  return TRUE;
}

static void
gprs_network_get_all_properties_cb(DBusPendingCall *pending, void *user_data)
{
  gprs_network_identifier *id = user_data;
  gprs_network *gprs = gprs_find_by_identifier(user_data);
  DBusMessage *reply = NULL;
  DBusMessageIter iter;

  ILOG_INFO("[GPRS] > gprs_network_get_all_properties_cb");

  if (gprs && pending)
  {
    reply = dbus_pending_call_steal_reply(pending);
    dbus_pending_call_unref(pending);

    if (reply)
    {
      if (dbus_message_get_type(reply) == DBUS_MESSAGE_TYPE_ERROR)
      {
        ILOG_WARN("[GPRS] property call returned '%s'",
                  dbus_message_get_error_name(reply));
      }
      else
      {
        ILOG_INFO("[GPRS] init the message value iterator %p", &iter);
        dbus_message_iter_init(reply, &iter);

        if (dbus_message_iter_get_arg_type(&iter) == DBUS_TYPE_ARRAY)
        {
          if (network_process_all_properties(&iter, gprs))
          {

            notify_observers(OBSERVER_TYPE_GPRS_13, OBSERVER_STATUS_OK,
                             gprs->priv);
            free_network_identifier(id);
            dbus_message_unref(reply);

            ILOG_INFO("[GPRS] < gprs_network_get_all_properties_cb");
            return;
          }
        }
      }
    }
  }

  if (gprs)
    notify_observers(OBSERVER_TYPE_GPRS_13, OBSERVER_STATUS_ERROR, gprs->priv);

  if (id)
    free_network_identifier(id);

  if (reply)
    dbus_message_unref(reply);
}

gboolean
context_get_all_properties(gprs_network *gprs)
{
  gprs_private *priv = gprs->priv;
  gprs_network_identifier *id;
  DBusMessage *message;
  const char *method = "com.nokia.csd.GPRS.Context";

  if (gprs->priv->observers[OBSERVER_TYPE_GPRS_PROPERTIES].active)
    return TRUE;

  id = create_network_identifier(gprs);
  message = dbus_message_new_method_call("com.nokia.csd.GPRS",
                                         gprs->path,
                                         "org.freedesktop.DBus.Properties",
                                         "GetAll");
  if (message)
  {
    if (dbus_message_append_args(message,
                                 DBUS_TYPE_STRING, &method,
                                 DBUS_TYPE_INVALID))
    {
      DBusPendingCall *mcall = icd_dbus_send_system_mcall(
            message, -1, gprs_network_get_all_properties_cb, id);

      if (mcall)
      {
        priv->observers[OBSERVER_TYPE_GPRS_PROPERTIES].mcall = mcall;
        priv->observers[OBSERVER_TYPE_GPRS_PROPERTIES].active = TRUE;
        dbus_message_unref(message);
        return TRUE;
      }
      else
        ILOG_ERR("[GPRS] could not send method call");
    }
    else
      ILOG_ERR("[GPRS] could not append args to the method call");

    dbus_message_unref(message);
  }
  else
    ILOG_ERR("[GPRS] could not create method call");

  free_network_identifier(id);

  return FALSE;
}

static enum observer_type
get_property_observer(const gchar *property)
{
  if (!property)
    return OBSERVER_TYPE_LAST;

  if (!strcmp(property, "APN"))
  {
    return OBSERVER_TYPE_APN;
  }
  if (!strcmp(property, "Password"))
    return OBSERVER_TYPE_PASSWORD;

  if (!strcmp(property, "Username"))
    return OBSERVER_TYPE_USERNAME;

  if (!strcmp(property, "PDPAddress"))
    return OBSERVER_TYPE_PDPADDRESS;

  return OBSERVER_TYPE_LAST;
}

static void
delete_property_item_2(property_item_2 *item)
{
  if (item)
  {
    g_free(item->val);
    item->val = NULL;
    g_free(item->key);
    item->key = NULL;
    g_free(item);
  }
}

static void
gprs_set_property_cb(DBusPendingCall *pending, void *user_data)
{
  property_item_2 *item = user_data;
  enum observer_type type = get_property_observer(item->key);
  DBusMessage *reply = NULL;
  enum observer_status status = OBSERVER_STATUS_ERROR;

  if (pending)
  {
    reply = dbus_pending_call_steal_reply(pending);
    dbus_pending_call_unref(pending);
  }

  if (pending && reply)
  {
    if (dbus_message_get_type(reply) == DBUS_MESSAGE_TYPE_ERROR)
    {
      ILOG_WARN("[GPRS] set_property call returned '%s'",
                dbus_message_get_error_name(reply));
    }
    else
      status = OBSERVER_STATUS_OK;

    dbus_message_unref(reply);
  }

  notify_observers(type, status, item->priv);
  delete_property_item_2(item);
}

static gboolean
gprs_set_property(gchar *key, gchar *val, gprs_network *gprs)
{
  observer *observer;
  property_item_2 *item;
  DBusMessage *message;

  observer = &gprs->priv->observers[get_property_observer(key)];

  if (observer->active)
    return TRUE;

  item = g_new0(property_item_2, 1);

  if (item)
  {
    item->priv = gprs->priv;
    item->key = g_strdup(key);
    item->val = g_strdup(val);
    message = dbus_message_new_method_call("com.nokia.csd.GPRS", gprs->path,
                                           "org.freedesktop.DBus.Properties",
                                           "Set");
    if (message)
    {
      const char *iface = "com.nokia.csd.GPRS.Context";

      if (dbus_message_append_args(message,
                                   DBUS_TYPE_STRING, &iface,
                                   DBUS_TYPE_STRING, &key,
                                   DBUS_TYPE_INVALID))
      {
        char contained_signature[2] = {DBUS_TYPE_STRING, 0};
        DBusPendingCall *mcall;
        DBusMessageIter sub;
        DBusMessageIter iter;

        dbus_message_iter_init_append(message, &iter);
        dbus_message_iter_open_container(&iter, DBUS_TYPE_VARIANT,
                                         contained_signature, &sub);
        dbus_message_iter_append_basic(&sub, DBUS_TYPE_STRING, &val);
        dbus_message_iter_close_container(&iter, &sub);
        mcall = icd_dbus_send_system_mcall(message, -1, gprs_set_property_cb,
                                           item);
        if (mcall)
        {
          observer->mcall = mcall;
          observer->active = TRUE;
          dbus_message_unref(message);
          return TRUE;
        }
        else
          ILOG_ERR("[GPRS] could not send method call");
      }
      else
        ILOG_ERR("[GPRS] could not append args to the method call");

      dbus_message_unref(message);
      delete_property_item_2(item);
    }
    else
      ILOG_ERR("[GPRS] could not create method call");
  }

  return FALSE;
}

static gboolean
csd_param_observer(enum observer_type type, enum observer_status status,
                   gprs_private *priv, gpointer user_data)
{
  gprs_network *gprs;
  gboolean rv = TRUE;
  enum icd_nw_status nw_status = ICD_NW_SUCCESS_NEXT_LAYER;
  const gchar *err_str = NULL;

  ILOG_INFO("[GPRS] > csd_param_observer for %i with status %i", type, status);

  if (status == OBSERVER_STATUS_ABORTED)
    return TRUE;

  gprs = gprs_find_by_id((const gchar *)user_data, priv);

  if (!gprs)
  {
    ILOG_INFO("[GPRS] no network for %s", (const char *)user_data);
    g_free(user_data);
    return FALSE;
  }

  g_free(user_data);

  if (!status )
  {
    switch (type)
    {
      case OBSERVER_TYPE_APN:
      {
        if (priv->observers[OBSERVER_TYPE_APN].active ||
            priv->observers[OBSERVER_TYPE_PASSWORD].active ||
            priv->observers[OBSERVER_TYPE_USERNAME].active)
        {
          return TRUE;
        }

        break;
      }
      case OBSERVER_TYPE_PASSWORD:
      {
        if (priv->observers[OBSERVER_TYPE_6].active ||
            priv->observers[OBSERVER_TYPE_PASSWORD].active ||
            priv->observers[OBSERVER_TYPE_USERNAME].active)
        {
          return TRUE;
        }

        break;
      }
      case OBSERVER_TYPE_USERNAME:
        if (priv->observers[OBSERVER_TYPE_APN].active ||
            priv->observers[OBSERVER_TYPE_6].active ||
            priv->observers[OBSERVER_TYPE_USERNAME].active)
        {
          return TRUE;
        }

        break;
      case OBSERVER_TYPE_PDPADDRESS:
      {
        if (priv->observers[OBSERVER_TYPE_APN].active ||
            priv->observers[OBSERVER_TYPE_PASSWORD].active ||
            priv->observers[OBSERVER_TYPE_6].active)
        {
          return TRUE;
        }

        break;
      }
      default:
        goto err_out;
    }
  }
  else
  {
err_out:
    nw_status = ICD_NW_ERROR;
    err_str = "Network error";
    rv = FALSE;
  }

  if (gprs->link_post_up)
  {
    gprs->link_post_up(nw_status, err_str, gprs->link_post_up_cb_token, NULL);
    gprs->link_post_up_cb_token = NULL;
    gprs->link_post_up = NULL;
  }
  else
    ILOG_ERR("[GPRS] ERROR: refused to call link_post_up_cb twice!");

  return rv;
}

gboolean
set_csd_parameters(gprs_network *gprs)
{
  gchar *network_id = g_strdup(gprs->network_id);
  gchar *ipv4_address = NULL;
  gchar *ipv4_type = NULL;

  if (!register_observer(OBSERVER_TYPE_APN, csd_param_observer, gprs->priv,
                         network_id, gprs->link_post_up_cb_token) ||
      !gprs_set_property("APN", gprs->accesspointname, gprs))
  {
    goto err_out;
  }

  if (gprs->password && *gprs->password)
  {
      if (!register_observer(OBSERVER_TYPE_PASSWORD, csd_param_observer,
                             gprs->priv, network_id,
                             gprs->link_post_up_cb_token) ||
          !gprs_set_property("Password", gprs->password, gprs))
      {
        goto err_out;
      }
  }

  if (gprs->username && *gprs->username)
  {
    if(!register_observer(OBSERVER_TYPE_USERNAME, csd_param_observer,
                          gprs->priv, network_id,
                          gprs->link_post_up_cb_token) ||
     !gprs_set_property("Username", gprs->username, gprs))
    {
      goto err_out;
    }
  }

  ipv4_type = icd_gconf_get_iap_string(gprs->network_id, "ipv4_type");
  ipv4_address = icd_gconf_get_iap_string(gprs->network_id, "ipv4_address");

  if (ipv4_type && !strcmp(ipv4_type, "STATIC") && ipv4_address)
  {
    if(!register_observer(OBSERVER_TYPE_PDPADDRESS, csd_param_observer,
                          gprs->priv, network_id,
                          gprs->link_post_up_cb_token) ||
       !gprs_set_property("PDPAddress", ipv4_address, gprs))
    {
      goto err_out;
    }
  }

  g_free(ipv4_type);
  g_free(ipv4_address);

  return TRUE;

err_out:
  g_free(ipv4_type);
  g_free(ipv4_address);
  abort_observation(csd_param_observer, gprs->priv,
                    gprs->link_post_up_cb_token);

  return FALSE;
}
