#include <icd/icd_gconf.h>
#include <icd/support/icd_log.h>

#include "gprs-private.h"
#include "observer.h"
#include "gprs.h"
#include "context.h"
#include "net.h"
#include "gconf.h"

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>

typedef struct
{
  const char *member;
  gboolean (*cb)(const char *, const char *, DBusMessage *, gpointer *user_data);
} dbus_signal;

static gboolean
csd_gprs_attach_failed(const char *member, const char *iface,
                       DBusMessage *message, gpointer *user_data)
{
  gprs_private *priv = *user_data;

  priv->net_attached = FALSE;

  return TRUE;
}

static gboolean
csd_gprs_attached(const char *member, const char *iface,
                  DBusMessage *message, gpointer *user_data)
{
  gprs_private *priv = *user_data;

  priv->net_attached = TRUE;

  return TRUE;
}

static gboolean
disconnect_timer_cb(gpointer user_data)
{
  gprs_private *priv = user_data;

  ILOG_ERR("[GPRS] Disconnect timeout expired, allowing connections");

  priv->disconnect_timer = 0;
  notify_observers(OBSERVER_TYPE_11, OBSERVER_STATUS_OK, priv);

  return FALSE;
}

static gboolean
csd_gprs_detached(const char *member, const char *iface,
                  DBusMessage *message, gpointer *user_data)
{
  gprs_private *priv = *user_data;

  priv->net_attached = FALSE;

  if (priv->suspension_timer)
  {
    ILOG_INFO("[GPRS] removing suspension timeout timer");
    g_source_remove(priv->suspension_timer);
    priv->suspension_timer = 0;
  }

  if (priv->networks)
  {
    if (!priv->disconnect_timer)
    {
      ILOG_ERR("[GPRS] Detached, asking for closing of the active contexts");

      close_all_contexts(priv, "Detached from the network",
                         priv->emergency_call ? NULL : "no_network");
    }
    else
    {
      ILOG_INFO("[GPRS] Detached, but the disconnect timer was on");

      g_source_remove(priv->disconnect_timer);
      disconnect_timer_cb(priv);
    }
  }

  priv->net_available = TRUE;

  return TRUE;
}

static gboolean
csd_gprs_available(const char *member, const char *iface, DBusMessage *message,
                   gpointer *user_data)
{
  gprs_private **priv = *user_data;

  ILOG_INFO("[GPRS] Network available");

  (*priv)->net_available = TRUE;

  if ((*priv)->suspension_timer)
  {
    ILOG_INFO("[GPRS] removing suspension timeout timer");
    g_source_remove((*priv)->suspension_timer);
    (*priv)->suspension_timer = 0;
  }

  return TRUE;
}

gboolean
datacounters_observer(enum observer_type type, enum observer_status status,
                      gprs_private *priv, gpointer user_data)
{
  if (status == OBSERVER_STATUS_ABORTED)
    return TRUE;
  else if (status == OBSERVER_STATUS_OK)
  {
    update_global_stats(priv, TRUE);
    return TRUE;
  }

  ILOG_ERR("[GPRS] Could not get baseline global counter values!");
  return FALSE;
}

static gboolean
csd_gprs_data_counters(const char *member, const char *iface,
                       DBusMessage *message, gpointer *user_data)
{
  gprs_private *priv = *user_data;
  gboolean rv = FALSE;
  dbus_uint64_t rx_bytes = 0;
  dbus_uint64_t tx_bytes = 0;

  if (dbus_message_get_args(message, NULL,
                            DBUS_TYPE_UINT64, &tx_bytes,
                            DBUS_TYPE_UINT64, &rx_bytes,
                            DBUS_TYPE_INVALID))
  {
    ILOG_INFO("[GPRS] Data Counter values: tx: %llu, rx: %llu",
              (unsigned long long)tx_bytes, (unsigned long long)rx_bytes);

    priv->tx_bytes = tx_bytes;
    priv->rx_bytes = rx_bytes;

    if (priv->dword50 == 0)
    {
      if (register_observer(OBSERVER_TYPE_DATACOUNTER, datacounters_observer,
                            priv, priv, (gpointer)message) &&
          mark_observer_active(priv, OBSERVER_TYPE_DATACOUNTER))
      {
        rv = TRUE;
      }
      else
        abort_observation(datacounters_observer, priv, (gpointer)message);
    }
    else
    {
      update_global_stats(priv, FALSE);
      rv = TRUE;
    }
  }
  else
    ILOG_WARN("[GPRS] could not get args from D-Bus message");

  return rv;
}

static dbus_signal
csd_gprs_handler[] =
{
  {"AttachFailed", csd_gprs_attach_failed},
  {"Attached", csd_gprs_attached},
  {"Detached", csd_gprs_detached},
  {"Available", csd_gprs_available},
/*  {"Suspended", csd_gprs_suspended}, */
  {"DataCounters", csd_gprs_data_counters},
  {NULL, NULL}
};

static gboolean
csd_gprs_context_connect_failed(const char *member, const char *iface,
                                DBusMessage *message, gpointer *user_data)
{
  gprs_network *gprs = *user_data;

  if (!gprs->ip_up_cb || !gprs->ip_up_cb_token)
    return FALSE;

  gprs->field_34 = 0;
  gprs->ip_up_cb(ICD_NW_ERROR, "[GPRS] network connection failure",
                 gprs->ip_up_cb_token, NULL);
  gprs->ip_up_cb_token = NULL;
  gprs->ip_up_cb = NULL;

  return TRUE;
}

gboolean
csd_gprs_context_connect_deleted(const char *member, const char *iface,
                                 DBusMessage *message, gpointer *user_data)
{
  gprs_network *gprs = *user_data;

  gprs->field_14 = 4;

  if (gprs->link_down_cb && gprs->link_down_cb_token )
  {
    gprs->priv->networks = g_slist_remove(gprs->priv->networks, gprs);

    if (!gprs->priv->disconnect_timer)
    {
      gprs->priv->disconnect_timer =
          g_timeout_add(4000, disconnect_timer_cb, gprs->priv);
    }

    gprs->link_down_cb(ICD_NW_SUCCESS, gprs->link_down_cb_token);
    gprs->link_down_cb_token = NULL;
    gprs->link_down_cb = NULL;
  }
  else
  {
    gprs->priv->networks = g_slist_remove(gprs->priv->networks, gprs);

    if (gprs->field_C)
    {
      if (gprs->ip_down_cb && gprs->ip_down_cb_token)
      {
        enum icd_nw_status nw_status;

        if (gprs->field_10)
          nw_status = ICD_NW_ERROR;
        else
          nw_status = ICD_NW_SUCCESS;

        gprs->ip_down_cb(nw_status, gprs->ip_down_cb_token);
        gprs->ip_down_cb_token = NULL;
        gprs->ip_down_cb = NULL;
      }
    }
    else
    {
      gprs->priv->close_fn(ICD_NW_ERROR, "Cellular context deleted",
                           gprs->network_type, gprs->network_attrs,
                           gprs->network_id);
    }
  }

  gprs_clear_data(gprs);
  g_free(gprs);

  return TRUE;
}

static gboolean
validate_interface(const char *iface)
{
  size_t i;
  size_t len = strlen(iface);

  if (!iface)
    return TRUE;

  for (i = 0; i < len; i++)
  {
    char c = iface[i];

    if (!isalpha(c) && !isdigit(c))
    {
      ILOG_ERR("[GPRS] ERROR: Interface '%s' did not pass validation!", iface);
      return FALSE;
    }
  }

  return TRUE;
}

static gboolean
validate_ip_address(const char *ip_address)
{
  size_t len = strlen(ip_address);
  size_t  i;

  if (ip_address)
    return TRUE;

  for (i = 0; i < len; i++)
  {
    char c = ip_address[i];

    if (c != ':' && c != '.' && !isdigit(c))
    {
      ILOG_ERR("[GPRS] ERROR: IP address '%s' did not pass validation!",
               ip_address);
      return 0;
    }
  }

  return TRUE;
}

static gboolean
connection_info_observer(enum observer_type type, enum observer_status status,
                         gprs_private *priv, gpointer user_data)
{
  gprs_network *gprs;
  const char *err_msg = NULL;

  if (status == OBSERVER_STATUS_ABORTED)
    return TRUE;

  if (!(gprs = gprs_find_by_id(user_data, priv)))
      ILOG_INFO("[GPRS] no network for %s", (char *)user_data);

  g_free(user_data);

  if (!gprs)
    return FALSE;

  if (status)
  {
    if (gprs->ip_up_cb)
    {
      gprs->ip_up_cb(ICD_NW_ERROR,
                     "asking for connection information from CSD failed",
                     gprs->ip_up_cb_token, NULL);
      gprs->ip_up_cb_token = NULL;
      gprs->ip_up_cb = NULL;
    }

    return FALSE;
  }

  if (!icd_gconf_get_iap_bool(gprs->network_id, "ipv4_autodns", TRUE) ||
      (!gprs->ip_dns1 && !gprs->ip_dns2))
  {
    gchar *dns1 = icd_gconf_get_iap_string(gprs->network_id, "ipv4_dns1");
    gchar *dns2 = icd_gconf_get_iap_string(gprs->network_id, "ipv4_dns2");

    if (dns1)
    {
      g_free(gprs->ip_dns1);
      gprs->ip_dns1 = dns1;
    }

    if (dns2)
    {
      g_free(gprs->ip_dns2);
      gprs->ip_dns2 = dns2;
    }
  }

  if (gprs->ip_address && (gprs->ip_dns1 || gprs->ip_dns2))
  {
    ILOG_ERR("[GPRS] Starting GPRS interface %s.\n", gprs->gprs_iface);

    if (validate_ip_address(gprs->ip_address) &&
        validate_ip_address(gprs->ip_dns1) &&
        validate_ip_address(gprs->ip_dns2) &&
        validate_interface(gprs->gprs_iface))
    {
      pid_t pid = fork();

      if (pid >= 0)
      {
        if (!pid)
        {
          execl("/etc/gprs/icd-gprs-up.sh", "/etc/gprs/icd-gprs-up.sh",
                gprs->gprs_iface, gprs->ip_address, gprs->ip_dns1,
                gprs->ip_dns2, NULL);
          exit(1);
        }

        gprs->pid_up = pid;
        gprs->priv->watch_fn(pid, gprs->priv->watch_fn_token);
      }
      else
        err_msg = "could not create the network interface";
    }
    else
      err_msg = "IP addresses did not pass validation";
  }
  else
    err_msg = "Could not determine IP address or DNS server address";

  if (err_msg)
  {
    if (gprs->ip_up_cb)
    {
      gprs->ip_up_cb(ICD_NW_ERROR, err_msg, gprs->ip_up_cb_token, NULL);
      gprs->ip_up_cb_token = NULL;
      gprs->ip_up_cb = NULL;
    }
    else
      ILOG_ERR("[GPRS] ERROR: refused to call ip_up_cb twice!");
  }

  return TRUE;
}

static gboolean
csd_gprs_context_connected(const char *member, const char *iface,
                           DBusMessage *message, gpointer *user_data)
{
  gprs_network *gprs = *user_data;

  ILOG_INFO("[GPRS] > connected");

  gprs->field_14 = 2;

  if (gprs->ip_up_cb && gprs->ip_up_cb_token)
  {
    if (register_observer(OBSERVER_TYPE_GPRS_13, connection_info_observer,
                          gprs->priv, g_strdup(gprs->network_id),
                          gprs->ip_up_cb_token) &&
        context_get_all_properties(gprs))
    {
      ILOG_INFO("[GPRS] < connected (TRUE)");
      return TRUE;
    }

    abort_observation(connection_info_observer, gprs->priv,
                      gprs->ip_up_cb_token);
    gprs->ip_up_cb(ICD_NW_ERROR, "[GPRS] network property request failure",
                   gprs->ip_up_cb_token, NULL);
    gprs->ip_up_cb_token = NULL;
    gprs->ip_up_cb = NULL;
  }

  ILOG_INFO("[GPRS] < connected (FALSE)");

  return FALSE;
}

static gboolean
csd_gprs_context_disconnected(const char *member, const char *iface,
                              DBusMessage *message, gpointer *user_data)
{
  gprs_network *gprs = *user_data;

  gprs->field_14 = 3;

  if (!gprs->ip_down_cb || !gprs->ip_down_cb_token)
  {
    gprs->field_C = 1;
    gprs->priv->close_fn(ICD_NW_ERROR, "Cellular context disconnected",
                         gprs->network_type, gprs->network_attrs,
                         gprs->network_id);
  }

  if (validate_interface(gprs->gprs_iface))
  {
    pid_t pid = fork();

    if (pid >= 0)
    {
      if (!pid)
      {
        execl("/etc/gprs/icd-gprs-down.sh", "/etc/gprs/icd-gprs-down.sh",
              gprs->gprs_iface, NULL);
        exit(1);
      }
      else
      {
        gprs->pid_down = pid;
        gprs->priv->watch_fn(pid, gprs->priv->watch_fn_token);
        return TRUE;
      }
    }
  }

  if (gprs->ip_down_cb && gprs->ip_down_cb_token)
  {
    gprs->ip_down_cb(ICD_NW_ERROR, gprs->ip_down_cb_token);
    gprs->ip_down_cb_token = NULL;
    gprs->ip_down_cb = NULL;
  }
  else
    gprs->field_10 = 1;

  return FALSE;
}

static dbus_signal
csd_gprs_context[] =
{
  {"Connected", csd_gprs_context_connected},
  {"ConnectFailed", csd_gprs_context_connect_failed},
  {"Disconnected", csd_gprs_context_disconnected},
  {"Deleted", csd_gprs_context_connect_deleted},
  {NULL, NULL}
};

static gboolean
phone_net_registration_status_change(const char *member, const char *iface,
                                     DBusMessage *message, gpointer *user_data)
{
  gprs_private *priv = *user_data;
  dbus_uint32_t network_country_code;
  dbus_uint32_t network_operator_code;
  dbus_uint32_t network_current_cell_id;
  dbus_uint16_t network_current_lac;
  guchar network_supported_services;
  guchar network_type;
  guchar network_reg_status;
  gboolean rv = FALSE;

  if (dbus_message_get_args(message, NULL,
                            DBUS_TYPE_BYTE, &network_reg_status,
                            DBUS_TYPE_UINT16, &network_current_lac,
                            DBUS_TYPE_UINT32, &network_current_cell_id,
                            DBUS_TYPE_UINT32, &network_operator_code,
                            DBUS_TYPE_UINT32, &network_country_code,
                            DBUS_TYPE_BYTE, &network_type,
                            DBUS_TYPE_BYTE, &network_supported_services,
                            DBUS_TYPE_INVALID))
  {
    guchar prev_reg_status = priv->net_reg_status;
    priv->net_reg_status = network_reg_status;
    network_request_check_connections(priv, prev_reg_status);

    if (priv->search_timeout_id && !is_searching(priv))
    {
      notify_observers(OBSERVER_TYPE_REG_STATUS, OBSERVER_STATUS_OK, priv);
      g_source_remove(priv->search_timeout_id);
      priv->search_timeout_id = 0;
    }

    rv = TRUE;
  }
  else
    ILOG_WARN("[GPRS] could not get args from D-Bus message");

  return rv;
}

static dbus_signal
phone_net_handler[] =
{
  {"registration_status_change", phone_net_registration_status_change},
  {NULL, NULL}
};

static gboolean
phone_sim_status(const char *member, const char *iface, DBusMessage *message,
                 gpointer *user_data)
{
  gprs_private *priv = *user_data;
  dbus_uint32_t sim_status;

  if (dbus_message_get_args(message, NULL,
                            DBUS_TYPE_UINT32, &sim_status,
                            DBUS_TYPE_INVALID))
  {
    priv->sim_status = sim_status;
    return TRUE;
  }

  ILOG_WARN("[GPRS] could not get args from D-Bus message");

  return FALSE;
}

static dbus_signal
phone_sim_handler[] =
{
  {"status", phone_sim_status},
  {NULL, NULL}
};

void
change_device_mode(const char *mode, gprs_private *priv)
{
  if (!strcmp(mode, "normal"))
    priv->online_status = 1;
  else if (!strcmp(mode, "flight") || !strcmp(mode, "offline"))
  {
    priv->online_status = 2;
    close_all_contexts(priv, "Entered offline mode", NULL);
  }
  else
    priv->online_status = 3;
}

static gboolean
mce_signal_sig_device_mode_ind(const char *member, const char *iface,
                               DBusMessage *message, gpointer *user_data)
{
  gprs_private *priv = *user_data;
  const char *mode;


  mode = 0;
  if ( dbus_message_get_args(message, 0,
                             DBUS_TYPE_STRING, &mode,
                             DBUS_TYPE_INVALID) )
  {
    change_device_mode(mode, priv);
    return TRUE;
  }

  ILOG_WARN("[GPRS] could not get args from D-Bus message");

  return FALSE;
}

static dbus_signal
mce_signal_handler[] =
{
  {"sig_device_mode_ind", mce_signal_sig_device_mode_ind},
  {NULL, NULL}
};

static gboolean
csd_call_server_status(const char *member, const char *iface,
                       DBusMessage *message, gpointer *user_data)
{
  gprs_private *priv = *user_data;
  dbus_bool_t emergency_call;
  dbus_bool_t unk;

  if (dbus_message_get_args(message, NULL,
                            DBUS_TYPE_BOOLEAN, &unk,
                            DBUS_TYPE_BOOLEAN, &emergency_call,
                            DBUS_TYPE_INVALID))
  {
    if (emergency_call)
    {
      ILOG_INFO("[GPRS] emergency call started");
      priv->emergency_call = TRUE;
    }
    else if (priv->emergency_call)
    {
      ILOG_INFO("[GPRS] emergency call ended");
      priv->emergency_call = FALSE;
    }

    return TRUE;
  }

  ILOG_WARN("[GPRS] could not get args from D-Bus message");

  return FALSE;
}

static dbus_signal
csd_call_handler[] =
{
  {"ServerStatus", csd_call_server_status},
  {NULL, NULL}
};


gboolean
dbus_handler_process_signal(const char *member, DBusMessage *message,
                            gpointer *priv)
{
  const char *iface = dbus_message_get_interface(message);
  dbus_signal *s;

  ILOG_INFO("[GPRS] Incoming signal '%s' on interface '%s'", member, iface);

  if (!strcmp(iface, "com.nokia.csd.GPRS"))
  {
    for (s = csd_gprs_handler; s->member; s++)
    {
        if (!strcmp(s->member, member) && s->cb)
          return s->cb(member, iface, message, priv);
    }

    ILOG_INFO("[GPRS] No suitable GPRS signal handler found");
    goto err_out;
  }

  if (!strcmp(iface, "com.nokia.csd.GPRS.Context"))
  {
    for (s = csd_gprs_context; s->member; s++)
    {
      if (!strcmp(s->member, member) && s->cb)
      {
        gprs_network *gprs =
            gprs_find_by_path(dbus_message_get_path(message), *priv);

        if (gprs)
          return s->cb(member, iface, message, (gpointer *)&gprs);
        else
        {
          ILOG_ERR("[GPRS] No suitable context data found");
          break;
        }
      }
    }

    ILOG_ERR("[GPRS] No suitable context signal handler found");
    goto err_out;
  }
  else if (!strcmp(iface, "Phone.Sim"))
  {
    for (s = phone_sim_handler; s->member; s++)
    {
      if (!strcmp(s->member, member) && s->cb)
        return s->cb(member, iface, message, priv);
    }
  }
  else if (!strcmp(iface, "Phone.Net"))
  {
    for (s = phone_net_handler; s->member; s++)
    {
      if (!strcmp(s->member, member) && s->cb)
        return s->cb(member, iface, message, priv);
    }
  }
  else if (!strcmp(iface, "com.nokia.csd.Call"))
  {
    for (s = csd_call_handler; s->member; s++)
    {
      if (!strcmp(s->member, member) && s->cb)
        return s->cb(member, iface, message, priv);
    }
  }
  else if (!strcmp(iface, "com.nokia.mce.signal"))
  {
    for (s = mce_signal_handler; s->member; s++)
    {
      if (!strcmp(s->member, member) && s->cb)
        return s->cb(member, iface, message, priv);
    }
  }
  else
  {
    ILOG_ERR("[GPRS] No match for the interface");
    goto err_out;
  }

  return TRUE;

err_out:
  ILOG_ERR("[GPRS] < process_signal (FALSE)");

  return FALSE;
}

DBusHandlerResult
network_sim_call_mce_signal_handler(DBusConnection *connection,
                                    DBusMessage *message, void *user_data)
{
  const char *iface = dbus_message_get_interface(message);

  if (dbus_message_get_type(message) == DBUS_MESSAGE_TYPE_SIGNAL &&
      (!strcmp(iface, "Phone.Net") || !strcmp(iface, "Phone.Sim") ||
       !strcmp(iface, "com.nokia.csd.Call") ||
       !strcmp(iface, "com.nokia.mce.signal")))
  {
    if (!dbus_handler_process_signal(dbus_message_get_member(message),
                                     message, &user_data))
    {
      ILOG_WARN("[GPRS] Signal processing failed");
    }
  }

  return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
}

DBusHandlerResult
gprs_context_signal_handler(DBusConnection *connection, DBusMessage *message, void *user_data)
{
  if (dbus_message_get_type(message) == DBUS_MESSAGE_TYPE_SIGNAL &&
      !strcmp(dbus_message_get_interface(message),
              "com.nokia.csd.GPRS.Context"))
  {
    if (!dbus_handler_process_signal(dbus_message_get_member(message), message,
                                     &user_data))
    {
      ILOG_WARN("[GPRS] Signal processing failed");
    }
  }

  return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
}

static dbus_signal *
dbus_handler_get_handler(const char *iface)
{
  dbus_signal *s = NULL;

  if (iface)
  {
    if (!strcmp(iface, "com.nokia.csd.GPRS"))
      s = csd_gprs_handler;
    else if (!strcmp(iface, "com.nokia.csd.GPRS.Context"))
      s = csd_gprs_context;
    else if (!strcmp(iface, "Phone.Net"))
      s = phone_net_handler;
    else if (!strcmp(iface, "Phone.Sim"))
      s = phone_sim_handler;
    else if (!strcmp(iface, "com.nokia.mce.signal"))
      s = mce_signal_handler;
    else if (!strcmp(iface, "com.nokia.csd.Call"))
      s = csd_call_handler;
  }

  return s;
}

gboolean
gprs_subscribe_to_signals(DBusConnection *connection, const char *iface,
                          DBusHandleMessageFunction cb, gpointer user_data)
{
  dbus_signal *sig;
  DBusError error;

  dbus_error_init(&error);

  if (!connection)
    return FALSE;

  sig = dbus_handler_get_handler(iface);

  if (!sig)
  {
    ILOG_ERR("[GPRS] tried to register unknown signal");
    return FALSE;
  }

  if (!dbus_connection_add_filter(connection, cb, user_data, NULL))
    ILOG_ERR("[GPRS] could not add signal filter");

  for(; sig->member; sig++)
  {
    char s[128];
    int res = snprintf(s, sizeof(s), "type='signal',interface='%s',member='%s'",
                       iface, sig->member);

    if (res == sizeof(s) || s < 0)
    {
      ILOG_ERR("[GPRS] signal match buffer creation error");
      return FALSE;
    }

    dbus_bus_add_match(connection, s, &error);

    if (dbus_error_is_set(&error))
    {
      ILOG_ERR("[GPRS] failed to add signal match rule");
      dbus_error_free(&error);
      return FALSE;
    }
  }

  dbus_error_free(&error);

  return TRUE;
}

gboolean
gprs_unsubscribe_to_signals(DBusConnection *connection, const char *iface,
                            DBusHandleMessageFunction cb, gpointer user_data)
{
  dbus_signal *sig;

  DBusError error;

  dbus_error_init(&error);

  if (!connection)
    return FALSE;

  sig = dbus_handler_get_handler(iface);

  if (!sig)
  {
    ILOG_ERR("[GPRS] tried to unregister unknown signal");
    return FALSE;
  }

  for (; sig->member; sig++)
  {
    char s[128];
    int res = snprintf(s, sizeof(s), "type='signal',interface='%s',member='%s'",
                       iface, sig->member);

    if (res == sizeof(s) || s < 0)
    {
      ILOG_ERR("[GPRS] signal match buffer creation error");
      return FALSE;
    }

    dbus_bus_remove_match(connection, s, &error);

    if (dbus_error_is_set(&error))
    {
      ILOG_ERR("[GPRS] failed to remove signal match rule");
      dbus_error_free(&error);
      return FALSE;
    }
  }

  dbus_connection_remove_filter(connection, cb, user_data);
  dbus_error_free(&error);

  return TRUE;
}
