#include <icd/support/icd_log.h>

#include <string.h>

#include "gprs-private.h"
#include "gconf.h"
#include "net.h"
#include "gprs.h"
#include "context.h"

static void
gprs_gconf_notify(GConfClient *client, guint cnxn_id, GConfEntry *entry,
                  gpointer user_data)
{
  gprs_private *priv = user_data;
  const char *key;
  GConfValue *val;

  if (!priv || !entry)
    return;

  key = gconf_entry_get_key(entry);

  if (!strcmp(key, ROAM_NTFY_PERIOD))
  {
    val = gconf_entry_get_value(entry);

    if (val)
    {
      priv->roaming_notification_period =
          g_ascii_strtoull(gconf_value_get_string(val), NULL, 10);
    }

    ILOG_INFO("[GPRS] the roaming notification period was set to %llu",
              (unsigned long long)priv->roaming_notification_period);
  }
  else if (!strcmp(key, HOME_NTFY_PERIOD))
  {
    val = gconf_entry_get_value(entry);

    if (val)
    {
      priv->home_notification_period =
          g_ascii_strtoull(gconf_value_get_string(val), NULL, 10);
    }

    ILOG_INFO("[GPRS] the home network notification period was set to %llu",
              (unsigned long long)priv->home_notification_period);
  }
}

gboolean
gprs_gconf_init(gprs_private *priv)
{
  gchar *s;
  GError *err = NULL;

  priv->gconf = gconf_client_get_default();

  if (!priv->gconf)
  {
    ILOG_ERR("[GPRS] ERROR: GConf error");
    return FALSE;
  }

  priv->cnxn = gconf_client_notify_add(priv->gconf,
                                       ICD_GCONF_NETWORK_MAPPING_GPRS,
                                       gprs_gconf_notify, priv, NULL, &err);
  if (err)
  {
    ILOG_ERR("[GPRS] GConf notification setting error: %s", err->message);
    goto err_out;
  }

  ILOG_DEBUG("[GPRS] GConf notification set for directory '%s'",
              ICD_GCONF_NETWORK_MAPPING_GPRS);

  s = gconf_client_get_string(priv->gconf, ROAM_NTFY_PERIOD, &err);

  if (err)
  {
      ILOG_ERR("[GPRS] GConf error while reading key ROAM_NTFY_PERIOD: %s",
               err->message);
      goto err_out;
  }

  if (s)
  {
    priv->roaming_notification_period = g_ascii_strtoull(s, NULL, 10);
    g_free(s);
  }
  else
    priv->roaming_notification_period = 10000000LL;

  s = gconf_client_get_string(priv->gconf, ROAM_LAST_NTFY, &err);

  if (err)
  {
    ILOG_ERR("[GPRS] GConf error while reading key ROAM_LAST_NTFY: %s",
               err->message);
    goto err_out;
  }

  if (s)
  {
    priv->roaming_last_notification = g_ascii_strtoull(s, NULL, 10);
    g_free(s);
  }

  s = gconf_client_get_string(priv->gconf, HOME_NTFY_PERIOD, &err);

  if (err)
  {
    ILOG_ERR("[GPRS] GConf error while reading key HOME_NTFY_PERIOD: %s",
               err->message);
    goto err_out;
  }

  if (s)
  {
    priv->home_notification_period = g_ascii_strtoull(s, NULL, 10);
    g_free(s);
  }
  else
    priv->home_notification_period = 10000000LL;

  s = gconf_client_get_string(priv->gconf, HOME_LAST_NTFY, &err);

  if (err)
  {
    ILOG_ERR("[GPRS] GConf error while reading key HOME_LAST_NTFY: %s",
             err->message);
    goto err_out;
  }

  if (s)
  {
    priv->home_last_notification = g_ascii_strtoull(s, NULL, 10);
    g_free(s);
  }

  return TRUE;

err_out:
  gconf_remove_notify(priv);

  return FALSE;
}

void
gconf_remove_notify(gprs_private *priv)
{
  if (priv->cnxn)
    gconf_client_notify_remove(priv->gconf, priv->cnxn);

  gconf_client_remove_dir(priv->gconf, ICD_GCONF_NETWORK_MAPPING_GPRS, NULL);
  g_object_unref(priv->gconf);
}

gboolean
gprs_roaming_enabled(gprs_private *priv)
{
  gboolean rv = FALSE;
  gboolean roaming_disabled;
  GError *error = NULL;

  roaming_disabled = gconf_client_get_bool(priv->gconf, ROAM_DISABLED, &error);

  if (error)
  {
    ILOG_DEBUG("[GPRS] GConf error: %s", error->message);
    g_error_free(error);
    roaming_disabled = TRUE;
  }
  else
    rv = roaming_disabled == FALSE;

  ILOG_DEBUG("[GPRS] Roaming allowed: %s", roaming_disabled ? "FALSE" : "TRUE");

  return rv;
}

gboolean
gconf_read_string_guint64(GConfClient *gconf, const gchar *key, guint64 *val)
{
  gboolean rv = FALSE;
  guint64 v = 0;
  GError *error = NULL;
  gchar *s = gconf_client_get_string(gconf, key, &error);

  if (!error)
  {
    if (s)
    {
      v = g_ascii_strtoull(s, NULL, 10);
      g_free(s);
    }

    rv = TRUE;
  }
  else
  {
    ILOG_ERR("[GPRS] GConf error while reading key %s: %s", key,
             error->message);
    g_error_free(error);
  }

  *val = v;

  return rv;
}

static gboolean
gconf_write_guint64_string(GConfClient *gconf, gchar *key, guint64 val)
{
  GError *error = NULL;
  gchar *s = g_strdup_printf("%"G_GUINT64_FORMAT, val);;
  gboolean rv = gconf_client_set_string(gconf, key, s, &error);;

  g_free(s);

  if (error)
  {
    ILOG_ERR("[GPRS] ERROR: GConf error while writing key %s: %s", key,
             error->message ? error->message : "NULL");

    g_error_free(error);
  }

  return rv;
}

gboolean
update_global_stats(gprs_private *priv, gboolean some_bool)
{
  guint64 combined;
  guint64 tx_now;
  guint64 tx_prev;
  guint64 rx_now;
  guint64 rx_prev;
  guint64 *tx_bytes;
  guint64 *rx_bytes;
  guint64 *notification_period;
  guint64 *last_notification;
  gchar *tx_key;
  gchar *rx_key;
  gchar *key;
  char *err_msg;
  guint64 home_rx_bytes;
  guint64 home_tx_bytes;
  guint64 roam_rx_bytes;
  guint64 roam_tx_bytes;
  guint64 rx_val;
  guint64 tx_val;

  tx_now = priv->tx_bytes;
  tx_prev = priv->tx_bytes_prev;
  rx_now = priv->rx_bytes;
  rx_prev = priv->rx_bytes_prev;

  if (!some_bool)
  {
    if (is_roaming(priv))
    {
      tx_key = ROAM_TX_BYTES;
      rx_key = ROAM_RX_BYTES;
      key = ROAM_LAST_NTFY;
      err_msg = "roaming_notification";
      tx_bytes = &priv->roaming_tx_bytes;
      rx_bytes = &priv->roaming_rx_bytes;
      notification_period = &priv->roaming_notification_period;
      last_notification = &priv->roaming_last_notification;
    }
    else
    {
      tx_key = HOME_TX_BYTES;
      rx_key = HOME_RX_BYTES;
      key = HOME_LAST_NTFY;
      err_msg = "home_notification";
      rx_bytes = &priv->home_rx_bytes;
      tx_bytes = &priv->home_tx_bytes;
      notification_period = &priv->home_notification_period;
      last_notification = &priv->home_last_notification;
    }

    ILOG_DEBUG("[GPRS] Current values: tx '%" G_GUINT64_FORMAT
               "' bytes + rx '%" G_GUINT64_FORMAT
               "' bytes = total '%" G_GUINT64_FORMAT
               "' bytes, last notification '%" G_GUINT64_FORMAT
               "', notification period '%" G_GUINT64_FORMAT "'", *tx_bytes,
               *rx_bytes, *rx_bytes + *tx_bytes, *last_notification,
               *notification_period);

    if (gconf_read_string_guint64(priv->gconf, tx_key, &tx_val) &&
        gconf_read_string_guint64(priv->gconf, rx_key, &rx_val))
    {
      guint64 tx_diff = 0;
      guint64 rx_diff = 0;
      gboolean v5 = FALSE;

      if (priv->tx_bytes_prev > priv->tx_bytes ||
          priv->rx_bytes_prev > priv->rx_bytes)
      {
        ILOG_ERR("[GPRS] ERROR: negative amount of data transferred");

        if (!priv->tx_bytes && !priv->rx_bytes)
        {
          priv->tx_bytes_prev = 0;
          priv->rx_bytes_prev = 0;
        }
      }
      else
      {
        tx_diff = tx_now - tx_prev;
        rx_diff = rx_now - rx_prev;

        if (*tx_bytes > tx_val || *rx_bytes > rx_val)
        {
          if (tx_val || rx_val)
          {
            ILOG_ERR(
                  "[GPRS] ERROR: global data counters are partially reseted?");
            *tx_bytes += tx_diff;
            *rx_bytes += rx_diff;
          }
          else
          {
            ILOG_INFO("[GPRS] global data counters were reset");
            *tx_bytes = tx_diff;
            *rx_bytes = rx_diff;
          }

          *last_notification = 0;
        }
        else
        {
          *tx_bytes += tx_diff;
          *rx_bytes += rx_diff;
        }

        v5 = 1;
        priv->tx_bytes_prev = priv->tx_bytes;
        priv->rx_bytes_prev = priv->rx_bytes;

        if (!gconf_write_guint64_string(priv->gconf, rx_key, *rx_bytes))
        {
          ILOG_ERR("[GPRS] ERROR: failed to write the RX bytes");
          if (v5)
            priv->rx_bytes_prev = *rx_bytes - rx_diff;
        }

        if (!gconf_write_guint64_string(priv->gconf, tx_key, *tx_bytes))
        {
          ILOG_ERR("[GPRS] ERROR: failed to write the TX bytes");

          if (v5)
            priv->tx_bytes_prev = *tx_bytes - tx_diff;
        }

        combined = *tx_bytes + *rx_bytes;

        if (*last_notification > combined)
          ILOG_ERR("[GPRS] ERROR: Last notification value is bigger than rx and tx bytes combined");

        if (*notification_period)
        {
          if (*notification_period <= combined - *last_notification)
          {
            ILOG_INFO("[GPRS] Showing the periodical notification");
            *last_notification = combined;

            if (!gconf_write_guint64_string(priv->gconf, key, combined))
              ILOG_ERR("[GPRS] ERROR: failed to set the last notification period to GConf");

            send_ui_request_without_reply(err_msg);
          }
        }

        gconf_client_suggest_sync(priv->gconf, NULL);
      }
    }
    else
    {
      ILOG_ERR("[GPRS] ERROR: GConf error in data counters");
      return FALSE;
    }
  }
  else
  {
    if (!gconf_read_string_guint64(priv->gconf, ROAM_TX_BYTES, &roam_tx_bytes) ||
        !gconf_read_string_guint64(priv->gconf, ROAM_RX_BYTES, &roam_rx_bytes) )
    {
      ILOG_ERR("[GPRS] ERROR: GConf error in roaming data counters");
      return FALSE;
    }

    if (!gconf_read_string_guint64(priv->gconf, HOME_TX_BYTES, &home_tx_bytes) ||
        !gconf_read_string_guint64(priv->gconf, HOME_RX_BYTES, &home_rx_bytes))
    {
      ILOG_ERR("[GPRS] ERROR: GConf error in home data counters");
      return FALSE;
    }

    priv->home_tx_bytes = home_tx_bytes;
    priv->home_rx_bytes = home_rx_bytes;
    priv->tx_bytes_prev = priv->prev_gprs_tx_bytes;
    priv->roaming_tx_bytes = roam_tx_bytes;
    priv->roaming_rx_bytes = roam_rx_bytes;
    priv->rx_bytes_prev = priv->prev_gprs_rx_bytes;
  }

  ILOG_DEBUG("[GPRS] Global statistics totals: home network '%" G_GUINT64_FORMAT
             "' tx bytes, '%" G_GUINT64_FORMAT "' rx bytes, roaming network '%"
             G_GUINT64_FORMAT "' tx bytes, '%" G_GUINT64_FORMAT "' rx bytes",
             priv->home_tx_bytes, priv->home_rx_bytes, priv->roaming_tx_bytes,
             priv->roaming_rx_bytes);

  return TRUE;
}
