#include <icd/support/icd_log.h>
#include <icd/network_api_defines.h>

#include <string.h>
#include <stdint.h>
#include <inttypes.h>

#include "gprs.h"

/**
 * @brief Helper function for comparing two strings where a NULL string is equal
 * to another NULL string
 *
 * @param a string A
 * @param b string B
 *
 * @return TRUE if equal, FALSE if unequal
 *
 */
inline static gboolean
string_equal(const char *a, const char *b)
{
  if (!a)
    return !b;

  if (b)
    return !strcmp(a, b);

  return FALSE;
}


void
free_network_identifier(gprs_network_identifier *id)
{
  if (!id)
    return;

  g_free(id->network_type);
  g_free(id->network_id);
  g_free(id);
}

gprs_network_identifier *
create_network_identifier(gprs_network *gprs)
{
  gprs_network_identifier *id = g_new0(gprs_network_identifier, 1);

  if (!id)
    return NULL;

  if ((id->network_type = g_strdup(gprs->network_type)) &&
      (id->network_id = g_strdup(gprs->network_id)))
  {
    id->priv = gprs->priv;
    id->network_attrs = gprs->network_attrs;
    return id;
  }

  free_network_identifier(id);

  return NULL;
}

gprs_network *
gprs_find(const gchar *network_type, const guint network_attrs,
          const gchar *network_id, gprs_private *priv)
{
  GSList *l;

  ILOG_DEBUG("[GPRS] > gprs_find");

  for (l = priv->networks; l; l = l->next)
  {
    gprs_network *gprs = l->data;

    if (gprs)
    {
      ILOG_INFO("[GPRS] checking %s/%d/%s", gprs->network_id,
                 gprs->network_attrs, gprs->network_type);

      if (gprs->network_attrs == network_attrs &&
          string_equal(gprs->network_type, network_type) &&
          string_equal(gprs->network_id, network_id))
      {
        ILOG_DEBUG("[GPRS] < gprs_find");
        return gprs;
      }
    }
    else
      ILOG_ERR("[GPRS] network data is NULL");
  }

  ILOG_DEBUG("[GPRS] < gprs_find (NULL)");
  return NULL;
}

gprs_network *
gprs_find_by_identifier(gprs_network_identifier *id)
{
  if (!id)
    return NULL;

  return
      gprs_find(id->network_type, id->network_attrs, id->network_id, id->priv);
}

gprs_network *
gprs_find_by_pid(pid_t pid, gprs_private *priv)
{
  GSList *l;

  ILOG_DEBUG("[GPRS] > gprs_find_by_pid: %i", pid);
  for (l = priv->networks; l; l = l->next)
  {
    gprs_network *gprs = l->data;

    if (gprs)
    {
      if (gprs->pid_up == pid || gprs->pid_down == pid)
      {
        ILOG_DEBUG("[GPRS] < gprs_find_by_pid: %ld, %p", (long)pid, gprs);
        return gprs;
      }
    }
    else
      ILOG_WARN("[GPRS] network data is NULL when searching for path");
  }

  ILOG_DEBUG("[GPRS] < gprs_find_by_pid (NULL)");

  return NULL;
}

gboolean
gprs_network_is_always_online(gprs_network *gprs_net)
{
  return gprs_net->network_attrs & ICD_NW_ATTR_ALWAYS_ONLINE;
}

gprs_network *
gprs_find_by_path(const gchar *path, gprs_private *priv)
{
  GSList *l;

  ILOG_DEBUG("[GPRS] > gprs_find_by_path: %s", path);

  for (l = priv->networks; l; l = l->next)
  {
    gprs_network *gprs = l->data;

    if (gprs)
    {
      if (!strcmp(path, gprs->path))
      {
        ILOG_DEBUG("[GPRS] < gprs_find_by_path: %s, %p", path, gprs);
        return gprs;
      }
    }
    else
      ILOG_WARN("[GPRS] network data is NULL when searching for path");

  }

  ILOG_DEBUG("[GPRS] < gprs_find_by_path (NULL)");

  return NULL;
}

gprs_network *
gprs_find_by_id(const gchar *network_id, gprs_private *priv)
{
  GSList *l;

  ILOG_DEBUG("[GPRS] > gprs_find_by_id: %s", network_id);

  for (l = priv->networks; l; l = l->next)
  {
    gprs_network *gprs = l->data;

    if (gprs)
    {
      ILOG_INFO("[GPRS] checking %s", gprs->network_id);

      if (string_equal(gprs->network_id, network_id))
      {
        ILOG_DEBUG("[GPRS] < gprs_find_by_id %p", gprs);
        return gprs;
      }
    }
    else
      ILOG_WARN("[GPRS] network data is NULL");
  }

  ILOG_DEBUG("[GPRS] < gprs_find_by_id (NULL)");

  return NULL;
}

void
gprs_clear_data(gprs_network *gprs)
{
  ILOG_DEBUG("[GPRS] > gprs_clear_data");

  g_free(gprs->network_type);
  gprs->network_type = NULL;
  gprs->network_attrs = 0;
  g_free(gprs->network_id);
  gprs->network_id = NULL;
  g_free(gprs->accesspointname);
  gprs->accesspointname = NULL;
  g_free(gprs->username);
  gprs->username = NULL;
  g_free(gprs->password);
  gprs->password = NULL;
  g_free(gprs->gprs_iface);
  gprs->gprs_iface = NULL;
  g_free(gprs->path);
  gprs->path = NULL;
  gprs->priv = NULL;
  g_free(gprs->ip_address);
  gprs->ip_address = NULL;
  g_free(gprs->ip_dns1);
  gprs->ip_dns1 = NULL;
  g_free(gprs->ip_dns2);
  gprs->ip_dns2 = NULL;

  ILOG_DEBUG("[GPRS] < gprs_clear_data");
}
