#include <icd/support/icd_dbus.h>
#include <icd/icd_gconf.h>
#include <icd/support/icd_log.h>
#include <osso-ic-dbus.h>

#include <errno.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>

#include "gprs-private.h"
#include "gprs.h"
#include "observer.h"
#include "gconf.h"
#include "context.h"
#include "dbus_handler.h"
#include "sim.h"
#include "net.h"

static gboolean
link_up_observer(enum observer_type type, enum observer_status status,
                 gprs_private *priv, gpointer user_data);

static void
gprs_delete_cb(DBusPendingCall *pending, void *user_data)
{
  gprs_network_identifier *gprs_net_id = user_data;
  gprs_network *gprs_net = gprs_find_by_identifier(gprs_net_id);

  ILOG_INFO("[GPRS] > gprs_delete_cb");

  free_network_identifier(gprs_net_id);

  if (pending)
  {
    DBusMessage *reply = dbus_pending_call_steal_reply(pending);

    dbus_pending_call_unref(pending);

    if (reply)
    {
      if (dbus_message_get_type(reply) != DBUS_MESSAGE_TYPE_ERROR)
      {
        dbus_message_unref(reply);
        ILOG_INFO("[GPRS] < gprs_delete_cb");
        return;
      }

      ILOG_WARN("[GPRS] 'Delete' call returned '%s'",
                dbus_message_get_error_name(reply));

      dbus_message_unref(reply);
    }
  }

  if (gprs_net)
  {
    if (gprs_net->link_down_cb)
    {
      gprs_net->link_down_cb(ICD_NW_ERROR, gprs_net->link_down_cb_token);
      gprs_net->link_down_cb_token = NULL;
      gprs_net->link_down_cb = NULL;
    }
    else
      ILOG_ERR("[GPRS] ERROR: refused to call link_down_cb twice!");
  }

  ILOG_INFO("[GPRS] < gprs_delete_cb (ERROR)");
}

static void
gprs_link_down(const gchar *network_type, const guint network_attrs,
               const gchar *network_id, const gchar *interface_name,
               icd_nw_link_down_cb_fn link_down_cb,
               const gpointer link_down_cb_token, gpointer *private)
{
  gprs_private *priv = *private;
  gprs_network *gprs_net =
      gprs_find(network_type, network_attrs, network_id, priv);
  DBusMessage *mcall;

  ILOG_INFO("[GPRS] nw link down");

  if (!gprs_net)
  {
    ILOG_INFO("[GPRS] no network for %s/%0x/%s", network_type, network_attrs,
              network_id);
    link_down_cb(ICD_NW_SUCCESS, link_down_cb_token);
    return;
  }

  gprs_net->link_down_cb = link_down_cb;
  gprs_net->link_down_cb_token = link_down_cb_token;
  mcall = dbus_message_new_method_call("com.nokia.csd.GPRS", gprs_net->path,
                                       "com.nokia.csd.GPRS.Context", "Delete");

  if (mcall)
  {
    if (icd_dbus_send_system_mcall(mcall, -1, gprs_delete_cb,
                                   create_network_identifier(gprs_net)))
    {
      dbus_message_unref(mcall);
      ILOG_INFO("[GPRS] < gprs_link_down");
      return;
    }

    ILOG_ERR("[GPRS] could not send 'Delete' mcall");
    dbus_message_unref(mcall);
  }
  else
    ILOG_ERR("[GPRS] could not create 'Delete' method call");

  gprs_net->link_down_cb_token = NULL;
  gprs_net->link_down_cb = NULL;
  link_down_cb(ICD_NW_ERROR, link_down_cb_token);
  ILOG_INFO("[GPRS] < gprs_link_down (ERROR)");
}

static void
gprs_connect_cb(DBusPendingCall *pending, void *user_data)
{
  gprs_network_identifier *gprs_net_id = user_data;
  gprs_network *gprs_net = gprs_find_by_identifier(gprs_net_id);

  free_network_identifier(gprs_net_id);

  if (pending)
  {
    DBusMessage *reply = dbus_pending_call_steal_reply(pending);

    dbus_pending_call_unref(pending);

    if (reply)
    {
      if (dbus_message_get_type(reply) != DBUS_MESSAGE_TYPE_ERROR)
      {
        dbus_message_unref(reply);
        return;
      }

      ILOG_WARN("[GPRS] 'Connect' call returned '%s'",
                dbus_message_get_error_name(reply));
      dbus_message_unref(reply);
    }
  }

  if (gprs_net)
  {
    if (gprs_net->ip_up_cb)
    {
      gprs_net->ip_up_cb(ICD_NW_ERROR,
                         "sending GPRS PDP context connection request failed",
                         gprs_net->ip_up_cb_token, NULL);
      gprs_net->ip_up_cb_token = NULL;
      gprs_net->ip_up_cb = NULL;
    }
    else
      ILOG_ERR("[GPRS] ERROR: refused to call ip_up_cb twice!");
  }

  ILOG_INFO("[GPRS] < gprs_connect_cb (ERROR)");
}

static void
gprs_ip_up(const gchar *network_type, const guint network_attrs,
           const gchar *network_id, const gchar *interface_name,
           icd_nw_ip_up_cb_fn ip_up_cb, gpointer ip_up_cb_token,
           gpointer *private)
{
  gprs_network *gprs_net;

  ILOG_DEBUG("[GPRS] nw ip up");

  gprs_net = gprs_find(network_type, network_attrs, network_id, *private);


  if (gprs_net)
  {
    DBusMessage *mcall;

    gprs_net->ip_up_cb = ip_up_cb;
    gprs_net->field_C = 0;
    gprs_net->ip_up_cb_token = ip_up_cb_token;

    mcall = dbus_message_new_method_call("com.nokia.csd.GPRS", gprs_net->path,
                                         "com.nokia.csd.GPRS.Context",
                                         "Connect");
    if (mcall)
    {
      if (icd_dbus_send_system_mcall(mcall, -1, gprs_connect_cb,
                                     create_network_identifier(gprs_net)))
      {
        dbus_message_unref(mcall);
        return;
      }
      else
      {
        dbus_message_unref(mcall);
        ILOG_ERR("[GPRS] could not send 'Connect' mcall");
      }
    }
    else
      ILOG_ERR("[GPRS] gprs could not create 'Connect' method call");

    gprs_net->ip_up_cb_token = NULL;
    gprs_net->ip_up_cb = NULL;
  }
  else
  {
    ILOG_INFO("[GPRS] no network for %s/%0x/%s", network_type, network_attrs,
              network_id);
  }

  ip_up_cb(ICD_NW_ERROR, "creating GPRS PDP context connection request failed",
           ip_up_cb_token, NULL);
}

static DBusHandlerResult
gprs_signal_handler(DBusConnection *connection, DBusMessage *message,
                    void *user_data)
{
  if (dbus_message_get_type(message) == DBUS_MESSAGE_TYPE_SIGNAL)
  {
    const char *iface = dbus_message_get_interface(message);

    if (iface && !strcmp(iface, "com.nokia.csd.GPRS"))
    {
      if (!dbus_handler_process_signal(dbus_message_get_member(message),
                                       message, user_data))
      {
        ILOG_WARN("[GPRS] Signal processing failed");
      }
    }
  }

  return TRUE;
}

static void
gprs_child_exit(pid_t pid, gint exit_value, gpointer *private)
{
  gprs_network *gprs = gprs_find_by_pid(pid, *private);

  if (!gprs)
    ILOG_INFO("[GPRS] GPRS has no network for pid %i", pid);
  else if (gprs->pid_up == pid)
  {
    if (!exit_value)
    {
      gprs->field_34 = TRUE;
      gprs->ts = time(0);
    }

    if (!gprs->ip_up_cb)
      ILOG_ERR("[GPRS] ERROR: refused to call ip_up_cb twice!");
    else
    {
      if (!exit_value)
        gprs->ip_up_cb(ICD_NW_SUCCESS, NULL, gprs->ip_up_cb_token, NULL);
      else
      {
        gprs->ip_up_cb(ICD_NW_ERROR,
                       "[GPRS] could not create the network interface",
                       gprs->ip_up_cb_token, NULL);
      }

      gprs->ip_up_cb_token = NULL;
      gprs->ip_up_cb = NULL;
    }

    gprs->pid_up = 0;
    return;
  }
  else if (gprs->pid_down == pid)
  {
    if (exit_value)
    {
      if (gprs->ip_down_cb && gprs->ip_down_cb_token)
        gprs->ip_down_cb(ICD_NW_ERROR, gprs->ip_down_cb_token);
      else
        gprs->field_10 = TRUE;
    }
    else
    {
      if (gprs->ip_down_cb && gprs->ip_down_cb_token)
      {
        gprs->field_34 = FALSE;
        gprs->ip_down_cb(ICD_NW_SUCCESS, gprs->ip_down_cb_token);
      }
    }

    gprs->pid_down = 0;
    gprs->ip_down_cb = NULL;
    gprs->ip_down_cb_token = NULL;
    return;
  }

  ILOG_ERR("[GPRS] Child process exit was not handled");
}

static void
gprs_network_destruct(gpointer *private)
{
  gprs_private *priv = *private;

  if (!priv)
    return;

  if (priv->networks)
  {
    ILOG_CRIT("[GPRS] still has connected networks");
  }

  g_free((gpointer)priv->imsi);
  gprs_unsubscribe_to_signals(priv->system_bus, "com.nokia.csd.GPRS",
                              gprs_signal_handler, priv);
  gprs_unsubscribe_to_signals(priv->system_bus, "com.nokia.csd.GPRS.Context",
                              gprs_context_signal_handler, priv);
  gprs_unsubscribe_to_signals(priv->system_bus, "Phone.Net",
                              network_sim_call_mce_signal_handler, priv);
  gprs_unsubscribe_to_signals(priv->system_bus, "Phone.Sim",
                              network_sim_call_mce_signal_handler, priv);
  gprs_unsubscribe_to_signals(priv->system_bus, "com.nokia.mce.signal",
                              network_sim_call_mce_signal_handler, priv);
  gprs_unsubscribe_to_signals(priv->system_bus, "com.nokia.csd.Call",
                              network_sim_call_mce_signal_handler, priv);
  dbus_connection_unref(priv->system_bus);
  priv->system_bus = NULL;
  gconf_remove_notify(priv);
  g_free(*private);
  *private = NULL;
}

static void
gprs_disconnect_cb(DBusPendingCall *pending, void *user_data)
{
  gprs_network *gprs;

  ILOG_INFO("[GPRS] > gprs_disconnect_cb");

  gprs = gprs_find_by_identifier(user_data);
  free_network_identifier(user_data);

  if (pending)
  {
    DBusMessage *reply = dbus_pending_call_steal_reply(pending);

    dbus_pending_call_unref(pending);

    if (reply)
    {
      if (dbus_message_get_type(reply) != DBUS_MESSAGE_TYPE_ERROR)
      {
        dbus_message_unref(reply);
        ILOG_INFO("[GPRS] < gprs_disconnect_cb");
        return;
      }

      ILOG_WARN("[GPRS] 'Disconnect' call returned '%s'",
                dbus_message_get_error_name(reply));
      dbus_message_unref(reply);
    }
  }

  if (gprs)
  {
    if (gprs->ip_down_cb)
    {
      gprs->ip_down_cb(ICD_NW_ERROR, gprs->ip_down_cb_token);
      gprs->ip_down_cb_token = NULL;
      gprs->ip_down_cb = NULL;
    }
    else
      ILOG_ERR("[GPRS] ERROR: refused to call ip_down_cb twice!");
  }

  ILOG_INFO("[GPRS] < gprs_disconnect_cb (ERROR)");
}

static void
gprs_ip_down(const gchar *network_type, const guint network_attrs,
             const gchar *network_id, const gchar *interface_name,
             icd_nw_ip_down_cb_fn ip_down_cb, gpointer ip_down_cb_token,
             gpointer *private)
{
  gprs_network *gprs;

  ILOG_DEBUG("[GPRS] nw ip down");

  gprs = gprs_find(network_type, network_attrs, network_id, *private);

  if (!gprs )
  {
    ILOG_INFO("[GPRS] has no network for %s/%0x/%s", network_type,
              network_attrs, network_id);
    goto err_out;
  }

  gprs->ip_down_cb = ip_down_cb;
  gprs->ip_down_cb_token = ip_down_cb_token;

  if (!gprs->field_C)
  {
    DBusMessage *mcall =
        dbus_message_new_method_call("com.nokia.csd.GPRS", gprs->path,
                                     "com.nokia.csd.GPRS.Context",
                                     "Disconnect");
    if (mcall)
    {
      if (icd_dbus_send_system_mcall(mcall, -1, gprs_disconnect_cb,
                                     create_network_identifier(gprs)))
      {
        dbus_message_unref(mcall);
        return;
      }

      ILOG_ERR("[GPRS] could not send 'Disconnect' mcall");
      dbus_message_unref(mcall);
    }
    else
      ILOG_ERR("[GPRS] could not create 'Disconnect' method call");

    gprs->ip_down_cb_token = NULL;
    gprs->ip_down_cb = NULL;
    goto err_out;
  }

  if (!gprs->pid_down)
  {
    enum icd_nw_status nw_status;

    if (gprs->field_10)
      nw_status = ICD_NW_ERROR;
    else
      nw_status = ICD_NW_SUCCESS;

    ip_down_cb(nw_status, ip_down_cb_token);
    gprs->ip_down_cb_token = NULL;
    gprs->ip_down_cb = NULL;
  }

  return;

err_out:
    ip_down_cb(ICD_NW_ERROR, ip_down_cb_token);
}

static void
gprs_stop_search(gpointer *private)
{
  gprs_private *priv = *private;

  ILOG_INFO("[GPRS] > gprs_stop_search");

  if (priv->search_cb)
    priv->search_cb(ICD_NW_SEARCH_COMPLETE, 0, 0, 0, 0, 0, 0, 0,
                    priv->search_cb_token);

  if (priv->read_iaps_id)
  {
    ILOG_DEBUG("[GPRS] stopping iap processing loop");
    g_source_remove(priv->read_iaps_id);
    priv->search_cb_token = NULL;
    priv->read_iaps_id = 0;
    priv->search_cb = NULL;

    ILOG_DEBUG("[GPRS] removing unprocessed iaps");

    while (priv->iap_dirs)
    {
      g_free(priv->iap_dirs->data);
      priv->iap_dirs->data = NULL;
      priv->iap_dirs = g_slist_delete_link(priv->iap_dirs, priv->iap_dirs);
    }
  }
  ILOG_INFO("[GPRS] < gprs_stop_search");
}

static gboolean
ipv4_stats_get(const char *iface,
               unsigned long long *rx_packets, unsigned long long *tx_packets,
               unsigned long long *rx_bytes, unsigned long long *tx_bytes)
{
  gboolean rv = FALSE;
  FILE *fp;
  char *col_pos;
  char *p;
  char s[256];
  unsigned long long tx_multicast;
  unsigned long long tx_compressed;
  unsigned long long tx_frame;
  unsigned long long tx_fifo;
  unsigned long long tx_drop;
  unsigned long long tx_errs;
  unsigned long long rx_multicast;
  unsigned long long rx_compressed;
  unsigned long long rx_frame;
  unsigned long long rx_fifo;
  unsigned long long rx_drop;
  unsigned long long rx_errs;

  fp = fopen("/proc/net/dev", "r");

  if (!fp)
    return FALSE;

  fgets(s, 256, fp);
  fgets(s, 256, fp);

  if (!strstr(s, "compressed"))
    goto out;

  do
  {
    if (!fgets(s, 256, fp))
      goto out;

    col_pos = strrchr(s, ':');

    p = s;

    while (*p == ' ')
      p++;
  }
  while (strncmp(p, iface, col_pos - p) || col_pos == (char *)-1);

  if (sscanf(col_pos + 1,
             "%llu %llu %llu %llu %llu %llu %llu %llu %llu %llu %llu %llu %llu %llu %llu %llu",
             rx_bytes, rx_packets, &rx_errs, &rx_drop, &rx_fifo, &rx_frame,
             &rx_compressed, &rx_multicast, tx_bytes, tx_packets, &tx_errs,
             &tx_drop, &tx_fifo, &tx_frame, &tx_compressed, &tx_multicast)
      == 16)
  {
    rv = TRUE;
  }

out:
  fclose(fp);

  return rv;
}

static void
gprs_ip_stats(const gchar *network_type, const guint network_attrs,
              const gchar *network_id, gpointer *private,
              icd_nw_ip_stats_cb_fn cb, const gpointer ip_stats_cb_token)
{
  gprs_private *priv = *private;
  gprs_network *gprs;

  if (!cb)
    return;

  gprs = gprs_find(network_type, network_attrs, network_id, priv);

  if (gprs)
  {
    unsigned long long tx_packets;
    unsigned long long rx_packets;
    unsigned long long tx_bytes;
    unsigned long long rx_bytes;

    if (ipv4_stats_get(gprs->gprs_iface, &rx_packets, &tx_packets, &rx_bytes,
                       &tx_bytes))
    {
      /* FIXME */
      /* fmg - Houston, we have a problem - on 64 bit system rx/tx bytes won't
         fit in guint, we'll have to change ICD2 and all the plugins :( */
      cb(ip_stats_cb_token, network_type, network_attrs, network_id,
         time(0) - gprs->ts, rx_bytes, tx_bytes);
      return;
    }
  }
  else
  {
    ILOG_INFO("[GPRS] has no network for %s/%0x/%s when statistics requested",
              network_type, network_attrs, network_id);
  }

  cb(ip_stats_cb_token, network_type, network_attrs, network_id, 0, 0, 0);

  ILOG_INFO("[GPRS] < gprs_ip_stats (FALSE)");
}

static gboolean
gprs_properties_observer(enum observer_type type, enum observer_status status,
                         gprs_private *priv, gpointer user_data)
{
  if (status == OBSERVER_STATUS_ABORTED)
    return TRUE;

  if (status)
  {
    notify_observers(OBSERVER_TYPE_DATACOUNTER, OBSERVER_STATUS_ERROR, priv);
    return FALSE;
  }

  priv->prev_gprs_tx_bytes = priv->gprs_tx_bytes;
  priv->prev_gprs_rx_bytes = priv->gprs_rx_bytes;
  priv->dword50 = TRUE;

  notify_observers(OBSERVER_TYPE_DATACOUNTER, OBSERVER_STATUS_OK, priv);

  return TRUE;
}

static gboolean
search_timeout_cb(void *user_data)
{
  gprs_private *priv = user_data;

  ILOG_ERR("[GPRS] search timeout expired, the network did not come up");
  notify_observers(OBSERVER_TYPE_REG_STATUS, OBSERVER_STATUS_ERROR, priv);

  return FALSE;
}

static gboolean
network_wait_until_registered(gprs_private *priv)
{
  ILOG_INFO("[GPRS] > network_wait_until_registered");

  if (priv->observers[OBSERVER_TYPE_2].active != TRUE)
  {
    if (!priv->search_timeout_id)
      priv->search_timeout_id = g_timeout_add(5000, search_timeout_cb, priv);

    priv->observers[OBSERVER_TYPE_2].mcall = NULL;
    priv->observers[OBSERVER_TYPE_2].active = TRUE;
  }

  return TRUE;
}

static DBusHandlerResult
roaming_cb(DBusConnection *connection, DBusMessage *message, void *user_data)
{
  gprs_private *priv = user_data;

  if (dbus_message_is_signal(message, "com.nokia.cellular_ui", "roaming"))
  {
    DBusError error;
    dbus_bool_t enabled;
    enum observer_status status;

    dbus_error_init(&error);

    if (dbus_message_get_args(message, &error,
                              DBUS_TYPE_BOOLEAN, &enabled,
                              DBUS_TYPE_INVALID))
    {

      if (enabled)
        status = OBSERVER_STATUS_OK;
      else
      {
        status = OBSERVER_STATUS_CANCEL;
        ILOG_INFO("[GPRS] roaming cancelled with status %d", status);
      }
    }
    else
    {
      ILOG_WARN("[GPRS] could not get args from 'roaming'");
      status = OBSERVER_STATUS_ERROR;
    }

    dbus_error_free(&error);

    icd_dbus_disconnect_system_bcast_signal("com.nokia.icd_ui", roaming_cb,
                                            priv, "member='roaming'");
    notify_observers(OBSERVER_TYPE_ICD_UI, status, priv);
  }

  return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
}

static void
show_roaming_dlg_cb(DBusPendingCall *pending, void *user_data)
{
  gprs_private *priv = user_data;

  if (pending)
  {
    DBusMessage *reply = dbus_pending_call_steal_reply(pending);

    dbus_pending_call_unref(pending);

    if(reply)
    {
      if (dbus_message_get_type(reply) == DBUS_MESSAGE_TYPE_ERROR)
      {
        ILOG_WARN("[GPRS] ask roaming returned '%s'",
                  dbus_message_get_error_name(reply));
        dbus_message_unref(reply);
      }
      else
      {
        ILOG_DEBUG("[GPRS] ask roaming successfully requested");
        dbus_message_unref(reply);
        return;
      }
    }
  }

  icd_dbus_disconnect_system_bcast_signal("com.nokia.icd_ui", roaming_cb, priv,
                                          "member='roaming'");
  notify_observers(OBSERVER_TYPE_ICD_UI, OBSERVER_STATUS_ERROR,
                   priv);
}

static gboolean
show_roaming_dlg(gprs_private *priv)
{
  gboolean rv = FALSE;

  if (priv->observers[OBSERVER_TYPE_REG_STATUS].active)
    return TRUE;

  if (icd_dbus_connect_system_bcast_signal("com.nokia.cellular_ui",
                                           roaming_cb, priv,
                                           "member='roaming'"))
  {
    DBusMessage *message =
        dbus_message_new_method_call("com.nokia.cellular_ui",
                                     "/com/nokia/cellular_ui",
                                     "com.nokia.cellular_ui",
                                     "show_roaming_dlg");
    if (message)
    {
      DBusPendingCall *mcall =
          icd_dbus_send_system_mcall(message, 10000, show_roaming_dlg_cb, priv);

      if (mcall)
      {
        priv->observers[OBSERVER_TYPE_REG_STATUS].mcall = mcall;
        priv->observers[OBSERVER_TYPE_REG_STATUS].active = TRUE;
        rv = TRUE;
      }
      else
        ILOG_ERR("[GPRS] could not send 'show_roaming_dlg' mcall");

      dbus_message_unref(message);
    }
    else
      ILOG_ERR("[GPRS] could not create 'show_roaming_dlg' method call");
  }
  else
    ILOG_ERR("[GPRS] could not connect to 'com.nokia.cellular_ui'");

  return rv;
}

static DBusHandlerResult
gprs_password_cb(DBusConnection *connection, DBusMessage *message,
                 void *user_data)
{
  gprs_private *priv = user_data;
  enum observer_status status = OBSERVER_STATUS_ERROR;
  DBusError error;
  dbus_bool_t ok;
  gchar *iap;
  gchar *password;
  gchar *username;

  if (!dbus_message_is_signal(message, "com.nokia.icd_ui", "passwd"))
    return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;

  dbus_error_init(&error);

  if (dbus_message_get_args(message, &error,
                            DBUS_TYPE_STRING, &username,
                            DBUS_TYPE_STRING, &password,
                            DBUS_TYPE_STRING, &iap,
                            DBUS_TYPE_BOOLEAN, &ok,
                            DBUS_TYPE_INVALID))
  {
    gprs_network *gprs = gprs_find_by_id(iap, priv);

    if (gprs)
    {
      if (gprs->network_id && iap && !strcmp(gprs->network_id, iap))
      {
        if (ok)
        {
          status = OBSERVER_STATUS_OK;
          g_free(gprs->username);
          gprs->username = g_strdup(username);
          g_free(gprs->password);
          gprs->password = g_strdup(password);
        }
        else
        {
          status = OBSERVER_STATUS_CANCEL;
          ILOG_INFO("[GPRS] password cancelled with status %d", status);
        }
      }
      else
      {
        ILOG_WARN("[GPRS] password reply for '%s' is not our network '%s'", iap,
                  gprs->network_id);
      }
    }
    else
      ILOG_INFO("[GPRS] no network for %s", iap);
  }
  else
    ILOG_INFO("[GPRS] could not get args from 'passwd'");

  dbus_error_free(&error);
  icd_dbus_disconnect_system_bcast_signal("com.nokia.icd_ui",
                                          gprs_password_cb, priv,
                                          "member='passwd'");
  notify_observers(OBSERVER_TYPE_5, status, priv);

  return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
}

static gboolean
addr_info_observer(enum observer_type type, enum observer_status status,
                   gprs_private *priv, gpointer user_data)
{
  gprs_network *gprs;
  gboolean rv = FALSE;

  if (status == OBSERVER_STATUS_ABORTED)
    return TRUE;

  gprs = gprs_find_by_id(user_data, priv);

  if (!gprs)
    ILOG_INFO("[GPRS] no network for %s", (char *)user_data);
  else if (status)
  {
    if (gprs->ip_addr_info_cb && gprs->addr_info_cb_token)
    {
        gprs->ip_addr_info_cb(gprs->addr_info_cb_token, gprs->network_type,
                              gprs->network_attrs, gprs->network_id, NULL, NULL,
                              NULL, NULL, NULL, NULL);
    }
  }
  else
  {
    if (gprs->ip_addr_info_cb && gprs->addr_info_cb_token)
    {
        gprs->ip_addr_info_cb(gprs->addr_info_cb_token, gprs->network_type,
                              gprs->network_attrs, gprs->network_id,
                              gprs->ip_address, "255.255.255.255", NULL,
                              gprs->ip_dns1, gprs->ip_dns2, NULL);
        gprs->addr_info_cb_token = NULL;
        gprs->ip_addr_info_cb = NULL;
    }

    rv = TRUE;
  }

  g_free(user_data);

  return rv;
}

static gboolean
process_gprs_properties(DBusMessage *reply, gprs_private *priv)
{
  DBusMessageIter sub4;
  DBusMessageIter sub2;
  DBusMessageIter sub1;
  DBusMessageIter iter;
  dbus_uint64_t value;

  ILOG_INFO("[GPRS] init the message value iterator %p", &iter);
  dbus_message_iter_init(reply, &iter);

  if (dbus_message_iter_get_arg_type(&iter) == DBUS_TYPE_ARRAY )
  {
    dbus_message_iter_recurse(&iter, &sub1);

    while (dbus_message_iter_get_arg_type(&sub1) == DBUS_TYPE_DICT_ENTRY)
    {
      char *key = NULL;

      dbus_message_iter_recurse(&sub1, &sub2);

      if (dbus_message_iter_get_arg_type(&sub2) != DBUS_TYPE_STRING)
        return FALSE;

      dbus_message_iter_get_basic(&sub2, &key);
      dbus_message_iter_next(&sub2);

      if (dbus_message_iter_get_arg_type(&sub2) != DBUS_TYPE_VARIANT)
        return FALSE;

      dbus_message_iter_recurse(&sub2, &sub4);

      if (!strcmp(key, "RxBytes"))
      {
        if (dbus_message_iter_get_arg_type(&sub4) != DBUS_TYPE_UINT64)
          return FALSE;

        dbus_message_iter_get_basic(&sub4, &value);
        ILOG_INFO("[GPRS] Read key '%s' and value '%llu'", key,
                  (unsigned long long)value);
        priv->gprs_rx_bytes = value;
      }
      else if (!strcmp(key, "TxBytes"))
      {
        if (dbus_message_iter_get_arg_type(&sub4) != DBUS_TYPE_UINT64)
          return FALSE;

        dbus_message_iter_get_basic(&sub4, &value);
        ILOG_INFO("[GPRS] Read key '%s' and value '%llu'", key,
                  (unsigned long long)value);
        priv->gprs_tx_bytes = value;
      }

      dbus_message_iter_next(&sub1);
    }
  }

  return TRUE;
}

static void
gprs_get_all_properties_cb(DBusPendingCall *pending, void *user_data)
{
  gprs_private *priv = user_data;
  DBusMessage *reply = NULL;
  enum observer_status status = OBSERVER_STATUS_ERROR;

  ILOG_INFO("[GPRS] > gprs_get_all_properties_cb");

  if (priv && pending)
  {
    reply = dbus_pending_call_steal_reply(pending);
    dbus_pending_call_unref(pending);

    if (reply)
    {
      if (dbus_message_get_type(reply) == DBUS_MESSAGE_TYPE_ERROR)
      {
        ILOG_WARN("[GPRS] property call returned '%s'",
                  dbus_message_get_error_name(reply));
      }
      else if (process_gprs_properties(reply, priv))
        status = OBSERVER_STATUS_OK;
    }
  }

  if (priv)
    notify_observers(OBSERVER_TYPE_GPRS_PROPERTIES, status, priv);

  if (reply)
    dbus_message_unref(reply);

  ILOG_INFO("[GPRS] < gprs_get_all_properties_cb");
}

static gboolean
gprs_get_all_properties(gprs_private *priv)
{
  gboolean rv = FALSE;
  DBusMessage *message;
  const char *method = "com.nokia.csd.GPRS";

  if (priv->observers[OBSERVER_TYPE_11].active)
    return TRUE;

  message = dbus_message_new_method_call("com.nokia.csd.GPRS",
                                         "/com/nokia/csd/gprs",
                                         "org.freedesktop.DBus.Properties",
                                         "GetAll");
  if (message)
  {
    if (dbus_message_append_args(message,
                                 DBUS_TYPE_STRING, &method,
                                 DBUS_TYPE_INVALID))
    {
      DBusPendingCall *mcall = icd_dbus_send_system_mcall(
            message, -1, gprs_get_all_properties_cb, priv);

      if (mcall)
      {
        priv->observers[OBSERVER_TYPE_11].mcall = mcall;
        priv->observers[OBSERVER_TYPE_11].active = TRUE;
        rv = TRUE;
      }
      else
        ILOG_ERR("[GPRS] could not send method call");
    }
    else
      ILOG_ERR("[GPRS] could not append args to the method call");

    dbus_message_unref(message);
  }
  else
    ILOG_ERR("[GPRS] could not create method call");

  return rv;
}

static void
gprs_ip_addr_info(const gchar *network_type, const guint network_attrs,
                  const gchar *network_id, gpointer *private,
                  icd_nw_ip_addr_info_cb_fn cb,
                  const gpointer addr_info_cb_token)
{
  gprs_private *priv = *private;
  gprs_network *gprs;
  gchar *net_id = NULL;

  if (!cb)
    return;

  gprs = gprs_find(network_type, network_attrs, network_id, priv);

  if (gprs)
  {
    if (gprs->field_34 && gprs->ip_address)
    {
      gprs->ip_addr_info_cb = cb;
      gprs->addr_info_cb_token = addr_info_cb_token;
      net_id = g_strdup(gprs->network_id);

      if (register_observer(OBSERVER_TYPE_GPRS_13, addr_info_observer, priv,
                            net_id, gprs->addr_info_cb_token))
      {
        if (context_get_all_properties(gprs))
          return;
      }

      abort_observation(addr_info_observer, priv, gprs->addr_info_cb_token);
    }
    else
      ILOG_INFO("[GPRS] not connected when address info requested");
  }
  else if ( icd_log_get_level() <= 1 )
  {
    ILOG_INFO("[GPRS] has no network for %s/%0x/%s when address info requested",
              network_type, network_attrs, network_id);
  }

  g_free(net_id);
  cb(addr_info_cb_token, network_type, network_attrs, network_id, NULL, NULL,
     NULL, NULL, NULL, NULL);
  ILOG_INFO("[GPRS] < gprs_ip_addr_info (FALSE)");
}

static gboolean
gprs_is_not_offline(gprs_private *priv)
{
  return priv->online_status != 0;
}

static gboolean
gprs_is_online(gprs_private *priv)
{
  ILOG_DEBUG("[GPRS] online status: '%i'", priv->online_status);

  return priv->online_status == 1;
}

static const char *
get_sim_status(gprs_private *priv)
{
  guint sim_status = priv->sim_status;


  if ( sim_status == 7 )
    return "no_pin";

  if (sim_status < 7 || sim_status - 9 > 1)
    return "no_sim";

  return "sim_locked";
}

static gboolean
is_registered(gprs_private *priv)
{
  ILOG_DEBUG("[GPRS] network reg status: '%i'", (int)priv->net_reg_status);

  return priv->net_reg_status <= 2;
}

static const char *
net_reg_status_to_error(gprs_private *priv)
{
  if (priv->net_reg_status == 6)
    return "no_sim";
  else
    return "no_network";
}

static gboolean
is_inited(gprs_private *priv)
{
  return priv->net_reg_status != 255;
}

static void
get_device_mode_cb(DBusPendingCall *pending, void *user_data)
{
  DBusMessage *reply;
  enum observer_status status = OBSERVER_STATUS_ERROR;

  if (pending)
  {
    reply = dbus_pending_call_steal_reply(pending);
    dbus_pending_call_unref(pending);
  }

  if (pending && reply)
  {
    if (dbus_message_get_type(reply) != DBUS_MESSAGE_TYPE_ERROR)
    {
      const char *mode = NULL;

      if (dbus_message_get_args(reply, NULL,
                                DBUS_TYPE_STRING, &mode,
                                DBUS_TYPE_INVALID))
      {
        change_device_mode(mode, user_data);
        status = OBSERVER_STATUS_OK;
      }
      else
        ILOG_WARN("[GPRS] could not get args from D-Bus message");
    }
    else
    {
      ILOG_WARN("[GPRS] ask device_state returned '%s'",
                dbus_message_get_error_name(reply));
    }

    dbus_message_unref(reply);
  }

  if (status != OBSERVER_STATUS_OK)
    change_device_mode("invalid", user_data);

  notify_observers(OBSERVER_TYPE_6, status, user_data);
}

static gboolean
get_device_mode(gprs_private *priv)
{
  DBusMessage *message;
  DBusPendingCall *mcall;

  if (priv->observers[OBSERVER_TYPE_5].active == TRUE)
    return TRUE;

  message = dbus_message_new_method_call("com.nokia.mce",
                                         "/com/nokia/mce/request",
                                         "com.nokia.mce.request",
                                         "get_device_mode");

  if (!message)
  {
    ILOG_ERR("[GPRS] could not create 'get_device_mode' method call");
    return FALSE;
  }

  mcall = icd_dbus_send_system_mcall(message, 10000, get_device_mode_cb, priv);

  if (mcall)
  {
    priv->observers[OBSERVER_TYPE_5].mcall = mcall;
    priv->observers[OBSERVER_TYPE_5].active = TRUE;
    dbus_message_unref(message);

    ILOG_DEBUG("[GPRS] sent mcall");
    return TRUE;
  }

  ILOG_ERR("[GPRS] could not send 'get_device_mode' mcall");
  dbus_message_unref(message);

  return FALSE;
}

static gboolean
ask_ui_for_roaming_permission(gprs_network *gprs, gprs_private *priv,
                              gpointer user_data)
{
  if (gprs_network_is_always_online(gprs) ||
      !register_observer(OBSERVER_TYPE_ICD_UI, link_up_observer,
                         priv, user_data, gprs->link_up_cb_token) ||
      !show_roaming_dlg(priv))
  {
    return FALSE;
  }

  return TRUE;
}

static gboolean
start_connection_cb(enum observer_type type, enum observer_status status,
                    gprs_private *priv, gpointer user_data)
{
  gprs_network *gprs;

  ILOG_INFO("[GPRS] > start_connection_cb");

  if (status == OBSERVER_STATUS_ABORTED)
    return TRUE;

  gprs = gprs_find_by_id((const gchar *)user_data, priv);

  if (gprs)
  {
    ILOG_INFO("[GPRS] no network for %s", (const char *)user_data);
    g_free(user_data);
    return FALSE;
  }

  if (status == OBSERVER_STATUS_OK)
  {
    if (gprs->gprs_iface)
    {
      if (gprs->link_up_cb)
      {
        gprs->link_up_cb(ICD_NW_SUCCESS_NEXT_LAYER, NULL, gprs->gprs_iface,
                         gprs->link_up_cb_token, NULL);
        gprs->link_up_cb_token = NULL;
        gprs->link_up_cb = NULL;
      }
      else
        ILOG_ERR("[GPRS] ERROR: refused to call link_up_cb twice!");
    }
    else
    {
      gprs->priv->networks = g_slist_remove(gprs->priv->networks, gprs);

      if (gprs->link_up_cb)
      {
        gprs->link_up_cb(ICD_NW_ERROR, "asking for interface from GPRS failed",
                         NULL, gprs->link_up_cb_token, NULL);
        gprs->link_up_cb_token = NULL;
        gprs->link_up_cb = NULL;
      }
      else
        ILOG_ERR("[GPRS] ERROR: refused to call link_up_cb twice!");

      gprs_clear_data(gprs);
      g_free(gprs);
    }

    ILOG_INFO("[GPRS] < start_connection_cb");
    g_free(user_data);
    return TRUE;
  }

  gprs->priv->networks = g_slist_remove(gprs->priv->networks, gprs);

  if (gprs->link_up_cb)
  {
    gprs->link_up_cb(ICD_NW_ERROR, "Asking for interface from CSD failed",
                     NULL, gprs->link_up_cb_token, NULL);
    gprs->link_up_cb_token = NULL;
    gprs->link_up_cb = NULL;
  }
  else
    ILOG_ERR("[GPRS] ERROR: refused to call link_up_cb twice!");

  gprs_clear_data(gprs);
  g_free(gprs);
  g_free(user_data);

  return FALSE;
}

static void
create_pdp_context_cb(DBusPendingCall *pending, void *user_data)
{
  gprs_network *gprs;
  DBusMessage *reply;

  gprs = gprs_find_by_identifier(user_data);
  free_network_identifier(user_data);

  ILOG_INFO("[GPRS] > create_pdp_context_cb");

  if (!gprs)
    return;

  if (pending)
  {
    reply = dbus_pending_call_steal_reply(pending);
    dbus_pending_call_unref(pending);
  }

  if (pending && reply)
  {
    if ( dbus_message_get_type(reply) == 3 )
    {

      ILOG_WARN("gprs 'Create' call returned '%s'",
                dbus_message_get_error_name(reply));

    }
    else
    {
      const gchar *path;

      if (!dbus_message_get_args(reply, 0, 'o', &path, 0))
        ILOG_WARN("[GPRS] could not get args from D-Bus message");
      else
      {
        gchar *net_id;

        g_free(gprs->path);
        gprs->path = g_strdup(path);
        dbus_message_unref(reply);
        ILOG_INFO("[GPRS] got the path %s for the new context", gprs->path);
        gprs->field_14 = 1;
        net_id = g_strdup(gprs->network_id);

        if (!register_observer(OBSERVER_TYPE_GPRS_13, start_connection_cb,
                               gprs->priv, net_id, gprs->link_up_cb_token) ||
            !context_get_all_properties(gprs))
        {
          ILOG_WARN("[GPRS] could not get network interface name from CSD");
          abort_observation(start_connection_cb, gprs->priv,
                            gprs->link_up_cb_token);
          g_free(net_id);
        }
        else
          ILOG_INFO("[GPRS] < create_pdp_context_cb (success)");

        return;
      }
    }

    dbus_message_unref(reply);
  }

  gprs->priv->networks = g_slist_remove(gprs->priv->networks, gprs);

  if (gprs->link_up_cb)
  {
    gprs->link_up_cb(ICD_NW_ERROR,
                     "sending GPRS PDP context creation request failed", NULL,
                     gprs->link_up_cb_token, NULL);
    gprs->link_up_cb_token = NULL;
    gprs->link_up_cb = NULL;
  }
  else
    ILOG_ERR("[GPRS] ERROR: refused to call link_up_cb twice!");

  gprs_clear_data(gprs);
  g_free(gprs);

  ILOG_INFO("[GPRS] < create_pdp_context_cb (failure)");

  return;
}

static void
create_pdp_context(gprs_network *gprs)
{
  DBusMessage *mcall;

  if (!gprs_network_is_always_online(gprs))
    send_ui_request_without_reply("req_autoconn_confirmation_dlg");

  mcall = dbus_message_new_method_call("com.nokia.csd.GPRS",
                                       "/com/nokia/csd/gprs",
                                       "com.nokia.csd.GPRS", "Create");
  if (mcall)
  {
    gprs_network_identifier *id = create_network_identifier(gprs);

    if (icd_dbus_send_system_mcall(mcall, -1, create_pdp_context_cb, id))
    {
      dbus_message_unref(mcall);
      return;
    }

    ILOG_ERR("[GPRS] could not send 'Create' mcall");
    free_network_identifier(id);
    dbus_message_unref(mcall);
  }
  else
    ILOG_ERR("[GPRS] could not create 'Create' method call");

  gprs->priv->networks = g_slist_remove(gprs->priv->networks, gprs);

  if (gprs->link_up_cb)
  {
    gprs->link_up_cb(ICD_NW_ERROR,
                     "creating GPRS PDP context creation request failed",
                     NULL, gprs->link_up_cb_token, NULL);
    gprs->link_up_cb_token = 0;
    gprs->link_up_cb = 0;
  }
  else
    ILOG_ERR("[GPRS] ERROR: refused to call link_up_cb twice!");

  gprs_clear_data(gprs);
  g_free(gprs);
}

static gboolean
link_up_observer(enum observer_type type, enum observer_status status,
                 gprs_private *priv, gpointer user_data)
{
  gprs_network *gprs;
  const gchar *err_str = "Network error";

  ILOG_INFO("[GPRS] > link_up_observer for %i with status %i", type, status);

  if (status == OBSERVER_STATUS_ABORTED)
    return TRUE;

  if (!(gprs = gprs_find_by_id(user_data, priv)))
  {
      ILOG_INFO("[GPRS] no network for %s", (const char *)user_data);
      g_free(user_data);
      return FALSE;
  }

  if (status)
  {
    if (type == OBSERVER_TYPE_6)
      err_str = "com.nokia.icd_ui.error.flight_mode";
    else
      err_str = "Network error";

    if (!gprs_network_is_always_online(gprs))
    {
      if (type == OBSERVER_TYPE_NET_STATUS)
        send_ui_request_without_reply(get_sim_status(priv));
      else if (type == OBSERVER_TYPE_2)
        send_ui_request_without_reply(net_reg_status_to_error(priv));
    }

    goto err_out;
  }

  switch (type)
  {
    case OBSERVER_TYPE_NET_STATUS:
    {
      if (priv->sim_status != 1)
        goto err_out;

      if (!is_inited(priv))
      {
        if (!register_observer(OBSERVER_TYPE_2, link_up_observer, priv,
                               user_data, gprs->link_up_cb_token) ||
            !network_request_registration_status(priv))
        {
          if (!gprs_network_is_always_online(gprs))
            send_ui_request_without_reply("no_network");

          err_str = "Could not query the network state";
          goto err_out;
        }

        return TRUE;
      }

      if (!is_registered(priv))
      {
        if (!gprs_network_is_always_online(gprs))
          send_ui_request_without_reply("no_network");

        err_str = "Not correctly registered to cellular network";
        goto err_out;
      }

      if (is_roaming(priv) && !gprs_roaming_enabled(priv) &&
          !ask_ui_for_roaming_permission(gprs, priv, user_data))
      {
        err_str = "Could not ask the UI for roaming permission";
        goto err_out;
      }

      if (priv->observers[OBSERVER_TYPE_5].active ||
          priv->observers[OBSERVER_TYPE_PDPADDRESS].active)
      {
        return TRUE;
      }

      goto LABEL_37;
    }
    case OBSERVER_TYPE_2:
    {
      if (!is_registered(priv))
        goto err_out;

      if (!is_roaming(priv) || gprs_roaming_enabled(priv))
      {
        if (priv->observers[OBSERVER_TYPE_5].active ||
            priv->observers[OBSERVER_TYPE_PDPADDRESS].active)
        {
          return TRUE;
        }

        goto LABEL_37;
      }

      if (!ask_ui_for_roaming_permission(gprs, priv, user_data))
      {
        err_str = "Could not ask the UI for roaming permission";
        goto err_out;
      }

      return TRUE;
    }
    case OBSERVER_TYPE_ICD_UI:
    {
      if (priv->observers[OBSERVER_TYPE_5].active ||
          priv->observers[OBSERVER_TYPE_SIM_IMSI].active ||
          priv->observers[OBSERVER_TYPE_PDPADDRESS].active)
      {
        return TRUE;
      }

      goto LABEL_37;
    }
    case OBSERVER_TYPE_6:
    {
      if (!gprs_is_online(priv))
      {
        err_str = "com.nokia.icd_ui.error.flight_mode";
        goto err_out;
      }

      if (priv->observers[OBSERVER_TYPE_NET_STATUS].active ||
          priv->observers[OBSERVER_TYPE_REG_STATUS].active ||
          priv->observers[OBSERVER_TYPE_SIM_IMSI].active ||
          priv->observers[OBSERVER_TYPE_PDPADDRESS].active)
      {
        return TRUE;
      }
LABEL_37:
      create_pdp_context(gprs);
      g_free(user_data);
      return TRUE;
    }
    case OBSERVER_TYPE_11:
    {
      if (priv->observers[OBSERVER_TYPE_NET_STATUS].active ||
          priv->observers[OBSERVER_TYPE_REG_STATUS].active ||
          priv->observers[OBSERVER_TYPE_SIM_IMSI].active ||
          priv->observers[OBSERVER_TYPE_5].active)
      {
        return TRUE;
      }

      goto LABEL_37;
    }
    default:
      goto err_out;
  }

  return TRUE;
err_out:
    ILOG_INFO("[GPRS] Network error from the observer, caused by %i status %i",
              type, status);

    if (gprs->link_up_cb)
    {
      gprs->link_up_cb(ICD_NW_ERROR, err_str, NULL, gprs->link_up_cb_token,
                       NULL);
      gprs->link_up_cb_token = NULL;
      gprs->link_up_cb = NULL;
    }
    else
      ILOG_ERR("[GPRS] ERROR: refused to call link_up_cb twice!");

    gprs->priv->networks = g_slist_remove(gprs->priv->networks, gprs);
    gprs_clear_data(gprs);
    g_free(gprs);
    g_free(user_data);
    return FALSE;

}

#define IAPS_PER_PASS 1

static gboolean
gprs_read_iaps(void *user_data)
{
  gprs_private *priv = user_data;
  int iap_cnt = 0;

  ILOG_INFO("[GPRS] > gprs_read_iaps");

  while (1)
  {
    const gchar *dir = priv->iap_dirs->data;
    gchar *type_key = g_strconcat(dir, "/type", NULL);
    gchar *type;

    if (type_key)
    {
      type = gconf_client_get_string(priv->gconf, type_key, NULL);
      g_free(type_key);

      if (type && !strcmp(type, "GPRS"))
      {
        gchar *key = g_strconcat(dir, "/", "sim_imsi", NULL);
        gchar *sim_imsi;
        gchar *name;

        if (key)
        {
          sim_imsi = gconf_client_get_string(priv->gconf, key, NULL);
          g_free(key);
        }
        else
          sim_imsi = NULL;

        key = g_strconcat(dir, "/", "name", NULL);

        if (key)
        {
          name = gconf_client_get_string(priv->gconf, key, NULL);
          g_free(key);
        }
        else
          name = NULL;

        if (sim_imsi && !strcmp(sim_imsi, priv->imsi))
        {
          gchar *iap = g_strrstr(dir, "/");

          if (iap)
          {
            char *iap_unescaped;

            iap++;

            ILOG_DEBUG("[GPRS] imsi (%s) match for IAP %s", sim_imsi, iap);
            iap_unescaped = gconf_unescape_key(iap, -1);

            if (!name)
            {
              ILOG_ERR("[GPRS] No UI name specified in GConf for iap '%s'!",
                       iap_unescaped);
            }

            priv->search_cb(ICD_NW_SEARCH_CONTINUE,
                            name ? name : "", type, 0x5000000,
                            iap_unescaped, ICD_NW_LEVEL_NONE, "GPRS",
                            ICD_NW_SUCCESS, priv->search_cb_token);
            g_free(iap_unescaped);
          }
        }
        else
          ILOG_DEBUG("[GPRS] no imsi match for IAP %s", dir);

        g_free(sim_imsi);
        g_free(name);
      }
    }
    else
      type = type_key;

    g_free(priv->iap_dirs->data);
    priv->iap_dirs->data = NULL;
    g_free(type);

    priv->iap_dirs = g_slist_delete_link(priv->iap_dirs, priv->iap_dirs);

    iap_cnt++;

    if (iap_cnt == IAPS_PER_PASS && priv->iap_dirs)
    {
      ILOG_DEBUG("[GPRS] read %d iaps but more to come, yielding", iap_cnt);
      break;
    }

    if (!priv->iap_dirs)
    {
      priv->read_iaps_id = 0;
      ILOG_WARN("[GPRS] Ending the scan (success)");
      priv->search_cb(ICD_NW_SEARCH_COMPLETE, 0, 0, 0, 0, 0, 0, 0,
                      priv->search_cb_token);
      ILOG_DEBUG("[GPRS] < gprs_read_iaps (FALSE)");
      return FALSE;
    }
  }

  if (!iap_cnt)
  {
    ILOG_DEBUG("[GPRS] reading all iaps from gconf");

    priv->iap_dirs = gconf_client_all_dirs(priv->gconf, ICD_GCONF_PATH, NULL);
  }

  if (priv->iap_dirs)
  {
    ILOG_DEBUG("[GPRS] < gprs_read_iaps (TRUE)");
    return TRUE;
  }

  ILOG_DEBUG("[GPRS] < gprs_read_iaps (FALSE)");

  return FALSE;
}

static void
gprs_link_up(const gchar *network_type, const guint network_attrs,
             const gchar *network_id, icd_nw_link_up_cb_fn link_up_cb,
             const gpointer link_up_cb_token, gpointer *private_)
{
  gprs_private *priv = *private_;
  gprs_network *gprs = g_new0(gprs_network, 1);
  gchar *net_id;
  signed int v14;
  const gchar *err_msg;
  guint sim_status;
  const char *sim_err_str = NULL;

  ILOG_DEBUG("[GPRS] > gprs_link_up");

  gprs->priv = priv;
  gprs->network_attrs = network_attrs;
  gprs->network_type = g_strdup(network_type);;
  gprs->network_id = g_strdup(network_id);
  gprs->link_up_cb_token = link_up_cb_token;
  gprs->link_up_cb = link_up_cb;
  gprs->field_14 = 0;
  gprs->path = g_strdup("");
  net_id = g_strdup(network_id);
  v14 = priv->emergency_call;

  if ( v14 )
  {
    err_msg = "Ongoing emergency call";
    goto LABEL_5;
  }

  if (gprs_is_not_offline(priv) )
  {
    if (!gprs_is_online(priv))
    {
      ILOG_WARN("[GPRS] device is not online");
      err_msg = "com.nokia.icd_ui.error.flight_mode";
LABEL_5:
      if ( icd_log_get_level() > 1 )
      {
LABEL_6:
        gprs->link_up_cb = 0;
        gprs->link_up_cb_token = 0;
        link_up_cb(ICD_NW_ERROR, err_msg, 0, link_up_cb_token, 0);
        abort_observation(link_up_observer, priv, link_up_cb_token);
        gprs_clear_data(gprs);
        g_free(gprs);
        g_free(net_id);
        return;
      }
LABEL_11:
      syslog(30, "[GPRS] Network error from the link_up");
      goto LABEL_6;
    }
  }
  else
  {
    if (register_observer(OBSERVER_TYPE_6, link_up_observer, priv, net_id,
                          link_up_cb_token) || !get_device_mode(priv))
    {
      err_msg = "com.nokia.icd_ui.error.flight_mode";
      if ( icd_log_get_level() > 1 )
        goto LABEL_6;
      goto LABEL_11;
    }
    v14 = 1;
  }

  sim_status = priv->sim_status;

  if (!sim_status)
  {
    if (register_observer(OBSERVER_TYPE_NET_STATUS, link_up_observer, priv,
                          net_id, link_up_cb_token))
    {
      if (sim_request_status(priv))
        goto LABEL_30;
    }

    sim_err_str = "no_sim";
    goto LABEL_17;
  }

  if (sim_status != 1)
  {
    sim_err_str = get_sim_status(priv);
    ILOG_WARN("[GPRS] SIM status error: %s", sim_err_str);
    goto LABEL_17;
  }

  if (!is_inited(priv))
  {
    if (!register_observer(OBSERVER_TYPE_2, link_up_observer, priv, net_id,
                           link_up_cb_token) ||
        !network_request_registration_status(priv))
    {
      goto LABEL_40;
    }
    goto LABEL_30;
  }

  if (!is_registered(priv))
  {
    ILOG_WARN("[GPRS] not correctly registered to network: %s",
              net_reg_status_to_error(priv));
    goto LABEL_17;
  }

  if (is_roaming(priv) && !gprs_roaming_enabled(priv))
  {
    if (gprs_network_is_always_online(gprs))
    {
      err_msg = "Network error";
      goto LABEL_5;
    }

    if (!register_observer(OBSERVER_TYPE_ICD_UI, link_up_observer,
                           priv, net_id, link_up_cb_token) ||
        !show_roaming_dlg(priv))
    {
      goto LABEL_19;
    }
LABEL_30:
    v14 = 1;
  }

  if (!priv->net_available)
  {
    ILOG_WARN("[GPRS] connection is suspended, link_up request denied");
LABEL_40:
    sim_err_str = "no_network";

LABEL_17:
    if (!gprs_network_is_always_online(gprs))
      send_ui_request_without_reply(sim_err_str);
    goto LABEL_19;
  }

  if (priv->disconnect_timer)
  {
    if (!register_observer(OBSERVER_TYPE_11, link_up_observer, priv, net_id,
                           link_up_cb_token) )
    {
LABEL_19:
      err_msg = "Network error";
      goto LABEL_5;
    }

    priv->networks = g_slist_append(priv->networks, gprs);
  }
  else
  {
    priv->networks = g_slist_append(priv->networks, gprs);

    if (!v14)
    {
      g_free(net_id);
      create_pdp_context(gprs);
      ILOG_DEBUG("[GPRS] < gprs nw link up");
    }
  }
}

static gboolean
start_search_observer(enum observer_type type, enum observer_status status,
                      gprs_private *priv, gpointer user_data)
{
  ILOG_INFO("[GPRS] > start_search_observer for %i with status %i", type,
            status);

  if (status == OBSERVER_STATUS_ABORTED)
    return TRUE;

  if (status == OBSERVER_STATUS_OK)
  {
    switch (type)
    {
      case OBSERVER_TYPE_SIM_IMSI:
      {
        if (!priv->observers[OBSERVER_TYPE_NET_STATUS].active &&
            !priv->observers[OBSERVER_TYPE_2].active &&
            !priv->observers[OBSERVER_TYPE_5].active &&
            !priv->observers[OBSERVER_TYPE_SIM_IMSI].active)
        {
            goto success;
        }

        return TRUE;
      }
      case OBSERVER_TYPE_NET_STATUS:
      {
        if (priv->sim_status != 1)
          break;

        if (priv->imsi_requested)
          return TRUE;

        if (priv->imsi)
        {
          if (priv->observers[OBSERVER_TYPE_NET_STATUS].active ||
              priv->observers[OBSERVER_TYPE_2].active)
          {
            return TRUE;
          }

          if (!priv->observers[OBSERVER_TYPE_5].active)
            goto success;

          return TRUE;
        }

        if (!register_observer(OBSERVER_TYPE_SIM_IMSI, start_search_observer,
                               priv, priv, priv->search_cb_token) ||
            !sim_request_imsi(priv))
        {
          break;
        }

        return TRUE;
      }
      case OBSERVER_TYPE_2:
      {
        if (is_registered(priv))
        {
          if (!priv->imsi_requested &&
              !priv->observers[OBSERVER_TYPE_5].active &&
              !priv->observers[OBSERVER_TYPE_SIM_IMSI].active)
          {
            goto success;
          }

          return TRUE;
        }

        if (is_searching(priv) &&
            register_observer(OBSERVER_TYPE_REG_STATUS, start_search_observer,
                              priv, priv, priv->search_cb_token) &&
            network_wait_until_registered(priv))
        {
          return TRUE;
        }

        break;
      }
      case OBSERVER_TYPE_REG_STATUS:
      {
        if (!is_registered(priv))
          break;

        if (!priv->imsi_requested &&
            !priv->observers[OBSERVER_TYPE_5].active &&
            !priv->observers[OBSERVER_TYPE_SIM_IMSI].active)
        {
          goto success;
        }

        return TRUE;
      }
      case OBSERVER_TYPE_6:
      {
        if (!gprs_is_online(priv))
          break;

        if (priv->observers[OBSERVER_TYPE_NET_STATUS].active ||
            priv->observers[OBSERVER_TYPE_2].active ||
            priv->imsi_requested ||
            priv->observers[OBSERVER_TYPE_SIM_IMSI].active)
        {
          return TRUE;
        }

        goto success;
      }
      default:
        break;
    }
  }

  ILOG_WARN("[GPRS] Ending the scan (failure)");
  priv->search_cb(ICD_NW_SEARCH_COMPLETE, 0, 0, 0, 0, 0, 0, 0,
                  priv->search_cb_token);
  ILOG_INFO("< start_search_observer (failure)");

  return FALSE;

success:
  priv->read_iaps_id = g_idle_add(gprs_read_iaps, priv);
  ILOG_INFO("[GPRS] < start_search_observer (success)");

  return TRUE;
}

static void
gprs_start_search(const gchar *network_type, guint search_scope,
                  icd_nw_search_cb_fn search_cb, const gpointer search_cb_token,
                  gpointer *private)
{
  gprs_private *priv = *private;

  priv->search_cb = search_cb;
  priv->search_cb_token = search_cb_token;

  ILOG_INFO("[GPRS] > gprs_start_search");

  if (priv->emergency_call)
    goto err_out;

  if (gprs_is_not_offline(priv))
  {
    if (!gprs_is_online(priv))
    {
      ILOG_WARN("[GPRS] device is not online");
      goto err_out;
    }
  }
  else if (!register_observer(OBSERVER_TYPE_6, start_search_observer, priv,
                              priv, search_cb_token) || !get_device_mode(priv))
  {
    goto err_out;
  }

  if (is_inited(priv))
  {
    if (!is_registered(priv))
    {
      ILOG_WARN("[GPRS] not correctly registered to network");

      if (!is_searching(priv) ||
          !register_observer(OBSERVER_TYPE_REG_STATUS, start_search_observer,
                             priv, priv, search_cb_token) ||
          !network_wait_until_registered(priv))
      {
        goto err_out;
      }
    }
  }
  else if (!register_observer(OBSERVER_TYPE_2, start_search_observer, priv,
                              priv, search_cb_token) ||
           !network_request_registration_status(priv))
  {
    goto err_out;
  }

  ILOG_INFO("[GPRS] reseting the IMSI status before scan");
  g_free(priv->imsi);
  priv->imsi = NULL;

  if (register_observer(OBSERVER_TYPE_SIM_IMSI,
                        start_search_observer, priv, priv, search_cb_token) &&
      sim_request_imsi(priv))
  {
    if (priv->net_available)
    {
      if (!priv->read_iaps_id )
        return;

      ILOG_WARN("[GPRS] module already searching: %u, search request denied",
                priv->read_iaps_id);
    }
    else
      ILOG_WARN("[GPRS] connection is suspended, search request denied");
  }
  else
    ILOG_ERR("[GPRS] module could not ask for IMSI");

err_out:
  ILOG_WARN("[GPRS] Ending the scan (failure)");
  search_cb(ICD_NW_SEARCH_COMPLETE, NULL, NULL, 0, NULL, 0, NULL, 0,
            search_cb_token);
  abort_observation(start_search_observer, priv, search_cb_token);
}

static void
show_passwd_dlg_cb(DBusPendingCall *pending, void *user_data)
{
  gprs_network_identifier *id = user_data;
  gprs_network *gprs = gprs_find_by_identifier(id);
  gprs_private *priv;
  DBusMessage *reply;

  if (gprs)
  {
    priv = gprs->priv;

    if (pending)
    {
      reply = dbus_pending_call_steal_reply(pending);
      dbus_pending_call_unref(pending);

      if (dbus_message_get_type(reply) == DBUS_MESSAGE_TYPE_ERROR)
      {
        ILOG_WARN("[GPRS] ask password returned '%s'",
                  dbus_message_get_error_name(reply));
        dbus_message_unref(reply);
      }
      else
      {
        ILOG_DEBUG("[GPRS] ask password successfully requested");
        priv->observers[OBSERVER_TYPE_ICD_UI].mcall = NULL;
        dbus_message_unref(reply);
        goto id_free; /*FIXME - why? */
      }
    }
  }
  else
    priv = id->priv;

  icd_dbus_disconnect_system_bcast_signal("com.nokia.icd_ui",
                                          gprs_password_cb, priv,
                                          "member='passwd'");
  notify_observers(OBSERVER_TYPE_5, OBSERVER_STATUS_ERROR, priv);

id_free:
  free_network_identifier(id);
}

static gboolean
show_passwd_dlg(gprs_network *gprs)
{
  char *empty = "";
  char **username;
  char **password;
  DBusMessage *message;

  if (!icd_dbus_connect_system_bcast_signal("com.nokia.icd_ui",
                                            gprs_password_cb,
                                            gprs->priv, "member='passwd'"))
  {
    ILOG_ERR("[GPRS] could not connect to 'com.nokia.icd_ui'");
    return FALSE;
  }

  message = dbus_message_new_method_call("com.nokia.icd_ui",
                                         "/com/nokia/icd_ui",
                                         "com.nokia.icd_ui",
                                         "show_passwd_dlg");
  if (message)
  {
    if (gprs->username)
      username = &gprs->username;
    else
      username = &empty;

    if (gprs->password)
      password = &gprs->password;
    else
      password = &empty;

    if (!dbus_message_append_args(message,
                                  DBUS_TYPE_STRING, username,
                                  DBUS_TYPE_STRING, password,
                                  DBUS_TYPE_STRING, &gprs->network_id,
                                  DBUS_TYPE_INVALID))
    {
      gprs_network_identifier *id = create_network_identifier(gprs);
      DBusPendingCall *mcall =
          icd_dbus_send_system_mcall(message, 10000, show_passwd_dlg_cb, id);

      if (mcall)
      {
        gprs->priv->observers[OBSERVER_TYPE_ICD_UI].active = TRUE;
        gprs->priv->observers[OBSERVER_TYPE_ICD_UI].mcall = mcall;
        dbus_message_unref(message);
        ILOG_DEBUG("[GPRS] sent mcall");
        return TRUE;
      }
      else
      {
        free_network_identifier(id);

        ILOG_ERR("[GPRS] could not send 'show_passwd_dlg' mcall");
      }
    }
    else
      ILOG_ERR("[GPRS] could not append args to 'show_passwd_dlg' method call");
  }
  else
    ILOG_ERR("[GPRS] could not create 'show_passwd_dlg' method call");

  if (message)
    dbus_message_unref(message);

  icd_dbus_disconnect_system_bcast_signal("com.nokia.icd_ui", gprs_password_cb,
                                          gprs->priv, "member='passwd'");

  return FALSE;
}

static gboolean
link_post_up_observer(enum observer_type type, enum observer_status status,
                      gprs_private *priv, gpointer user_data)
{
  gprs_network *gprs;
  const gchar *err_str;

  if (status == OBSERVER_STATUS_ABORTED)
    return TRUE;

  gprs = gprs_find_by_id(user_data, priv);

  if (!gprs)
  {
    ILOG_INFO("[GPRS] no network for %s", (const char *)user_data);
    return FALSE;
  }

  g_free(user_data);

  if (status == OBSERVER_STATUS_OK)
  {
    if (!set_csd_parameters(gprs))
      err_str = "com.nokia.icd.error.gateway_error";
    else
      return TRUE;
  }
  else
  {
    if (status == OBSERVER_STATUS_CANCEL)
      err_str = "com.nokia.icd.error.ppp_authentication_failed";
    else
      err_str = "com.nokia.icd.error.system_error";
  }

  if (gprs->link_post_up)
  {
    gprs->link_post_up(ICD_NW_ERROR, err_str, gprs->link_post_up_cb_token,
                       NULL);
    gprs->link_post_up_cb_token = NULL;
    gprs->link_post_up = NULL;

  }
  else
    ILOG_ERR("[GPRS] ERROR: refused to call link_post_up_cb twice!");

  return FALSE;
}

static void
gprs_link_post_up(const gchar *network_type, const guint network_attrs,
                  const gchar *network_id, const gchar *interface_name,
                  icd_nw_link_post_up_cb_fn link_post_up,
                  const gpointer link_post_up_cb_token, gpointer *private_)
{
  gprs_private *priv = *private_;
  gprs_network *gprs;

  ILOG_DEBUG("[GPRS] nw link post up");

  gprs = gprs_find(network_type, network_attrs, network_id, priv);

  if (!gprs)
  {
    ILOG_ERR("[GPRS] has no network for %s/%0x/%s", network_type,
              network_attrs, network_id);
    link_post_up(ICD_NW_ERROR, "GPRS did not find network for link_post_up",
                 link_post_up_cb_token, NULL);
    abort_observation(link_post_up_observer, priv, link_post_up_cb_token);
    return;
  }

  gprs->link_post_up_cb_token = link_post_up_cb_token;
  gprs->link_post_up = link_post_up;
  g_free(gprs->accesspointname);
  gprs->accesspointname =
      icd_gconf_get_iap_string(gprs->network_id, "gprs_accesspointname");

  if (gprs->accesspointname)
  {
    g_free(gprs->username);
    gprs->username =
        icd_gconf_get_iap_string(gprs->network_id, "gprs_username");
    g_free(gprs->password);
    gprs->password =
        icd_gconf_get_iap_string(gprs->network_id, "gprs_password");

    if (gprs_network_is_always_online(gprs))
    {
      if (set_csd_parameters(gprs))
        return;
      else
      {
        ILOG_ERR(
              "[GPRS] could not set CSD parameters, going to error handling");
      }
    }
    else
    {
      if (register_observer(OBSERVER_TYPE_5, link_post_up_observer, priv,
                            g_strdup(network_id), link_post_up_cb_token)
          && show_passwd_dlg(gprs))
      {
        return;
      }
    }
  }

  link_post_up(ICD_NW_ERROR, "GPRS did not find network for link_post_up",
               link_post_up_cb_token, NULL);
  abort_observation(link_post_up_observer, priv, link_post_up_cb_token);
  gprs->link_post_up = 0;
  gprs->link_post_up_cb_token = 0;
  g_free(gprs->accesspointname);
  gprs->accesspointname = NULL;
}

gboolean
icd_nw_init(struct icd_nw_api *network_api, icd_nw_watch_pid_fn watch_fn,
            gpointer watch_fn_token, icd_nw_close_fn close_fn,
            icd_nw_status_change_fn status_change_fn, icd_nw_renew_fn renew_fn)
{
  gprs_private *priv = g_new0(gprs_private, 1);
  DBusError error;

  ILOG_INFO("[GPRS] new gprs_data: %p", priv);

  network_api->version = ICD_NW_MODULE_VERSION;
  network_api->link_up = gprs_link_up;
  network_api->link_down = gprs_link_down;
  network_api->link_post_up = gprs_link_post_up;
  network_api->ip_up = gprs_ip_up;
  network_api->ip_down = gprs_ip_down;
  network_api->start_search = gprs_start_search;
  network_api->stop_search = gprs_stop_search;
  network_api->ip_addr_info = gprs_ip_addr_info;
  network_api->ip_stats = gprs_ip_stats;
  network_api->search_lifetime = 30;
  network_api->search_interval = 20;
  network_api->network_destruct = gprs_network_destruct;
  network_api->child_exit = gprs_child_exit;

  priv->close_fn = close_fn;
  priv->watch_fn_token = watch_fn_token;
  priv->watch_fn = watch_fn;
  priv->rx_bytes = 0LL;
  priv->prev_gprs_tx_bytes = 0LL;
  priv->prev_gprs_rx_bytes = 0LL;
  priv->home_tx_bytes = 0LL;
  priv->home_rx_bytes = 0LL;
  priv->roaming_tx_bytes = 0LL;
  priv->roaming_rx_bytes = 0LL;
  priv->tx_bytes_prev = 0LL;
  priv->rx_bytes_prev = 0LL;
  priv->gprs_tx_bytes = 0LL;
  priv->gprs_rx_bytes = 0LL;
  priv->tx_bytes = 0LL;
  priv->read_iaps_id = 0;
  priv->net_reg_status = 255;
  priv->imsi = 0;
  priv->suspension_timer = 0;
  priv->search_timeout_id = 0;
  priv->disconnect_timer = 0;
  priv->sim_status = 0;
  priv->emergency_call = 0;
  priv->dword50 = 0;
  priv->net_available = TRUE;

  dbus_error_init(&error);
  priv->system_bus = dbus_bus_get(DBUS_BUS_SYSTEM, &error);

  if (dbus_error_is_set(&error))
  {
    ILOG_ERR("[GPRS] failed to get system bus");
    dbus_error_free(&error);
    g_free(priv);
    return FALSE;
  }

  if (!gprs_gconf_init(priv))
    goto err_dbus;

  if (!gprs_subscribe_to_signals(priv->system_bus, "com.nokia.csd.GPRS",
                                 gprs_signal_handler, priv))
  {
    ILOG_ERR("[GPRS] module could not register CSD signal reception for GPRS");
    goto err_gconf;
  }


  if (!gprs_subscribe_to_signals(priv->system_bus, "com.nokia.csd.GPRS.Context",
                                 gprs_context_signal_handler, priv))
  {
    ILOG_ERR(
      "[GPRS] module could not register CSD signal reception for GPRS context");
    goto err_gconf;
  }

  if (!gprs_subscribe_to_signals(priv->system_bus, "Phone.Net",
                                 network_sim_call_mce_signal_handler, priv) ||
      !gprs_subscribe_to_signals(priv->system_bus, "Phone.Sim",
                                 network_sim_call_mce_signal_handler, priv))
  {
    ILOG_ERR(
          "[GPRS] module could not register CSD signal reception for network");
    goto err_gconf;
  }

  if (!gprs_subscribe_to_signals(priv->system_bus, "com.nokia.mce.signal",
                                 network_sim_call_mce_signal_handler, priv))
  {

    ILOG_ERR("[GPRS] module could not register MCE signal reception");
    goto err_gconf;
  }

  if (!gprs_subscribe_to_signals(priv->system_bus, "com.nokia.csd.Call",
                                 network_sim_call_mce_signal_handler, priv))
  {
    ILOG_ERR("[GPRS] module could not register CSD call signal reception");
    goto err_gconf;
  }

  if (!register_observer(OBSERVER_TYPE_DATACOUNTER, datacounters_observer, priv,
                         priv, network_api) ||
      !mark_observer_active(priv, OBSERVER_TYPE_DATACOUNTER))
  {
    ILOG_ERR("[GPRS] couldn't register for asking global counters");
    abort_observation(datacounters_observer, priv, network_api);
    goto err_gconf;
  }

  if (register_observer(OBSERVER_TYPE_GPRS_PROPERTIES, gprs_properties_observer,
                        priv, priv, network_api) &&
      gprs_get_all_properties(priv))
  {

    network_api->private = priv;

    return TRUE;
  }

  ILOG_ERR("[GPRS] register for asking for GPRS properties");
  abort_observation(gprs_properties_observer, priv, network_api);

err_gconf:
  gconf_client_notify_remove(priv->gconf, priv->cnxn);
  g_object_unref(priv->gconf);

err_dbus:
  dbus_connection_close(priv->system_bus);
  g_free(priv);
  return FALSE;
}
