#include <icd/support/icd_dbus.h>
#include <icd/support/icd_log.h>

#include "gprs-private.h"
#include "observer.h"
#include "gprs.h"
#include "net.h"
#include "gconf.h"
#include "context.h"

enum net_registration_status
{
    NETWORK_REG_STATUS_HOME = 0x00, // CS is registered to home network
    NETWORK_REG_STATUS_ROAM,        // CS is registered to some other network than home network
    NETWORK_REG_STATUS_ROAM_BLINK,  // CS is registered to non-home system in a non-home area ('a' or 'b' area)
    NETWORK_REG_STATUS_NOSERV,      // CS is not in service
    NETWORK_REG_STATUS_NOSERV_SEARCHING,    // CS is not in service, but is currently searching for service
    NETWORK_REG_STATUS_NOSERV_NOTSEARCHING, // CS is not in service and it is not currently searching for service
    NETWORK_REG_STATUS_NOSERV_NOSIM,        // CS is not in service due to missing SIM or missing subscription
    NETWORK_REG_STATUS_POWER_OFF = 0x08,    // CS is in power off state
    NETWORK_REG_STATUS_NSPS,                // CS is in No Service Power Save State (currently not listening to any cell)
    NETWORK_REG_STATUS_NSPS_NO_COVERAGE,    // CS is in No Service Power Save State (CS is entered to this state because there is no network coverage)
    NETWORK_REG_STATUS_NOSERV_SIM_REJECTED_BY_NW // CS is not in service due to missing subscription
};

static gboolean
reg_status_is_roaming(guchar net_reg_status)
{
  return net_reg_status == NETWORK_REG_STATUS_ROAM ||
      net_reg_status == NETWORK_REG_STATUS_ROAM_BLINK;
}

gboolean
is_roaming(gprs_private *priv)
{
  return reg_status_is_roaming(priv->net_reg_status);
}

gboolean
is_searching(gprs_private *priv)
{
  ILOG_DEBUG("[GPRS] network reg status: '%i'", (int)priv->net_reg_status);

  return priv->net_reg_status == NETWORK_REG_STATUS_NOSERV_SEARCHING;
}

void
network_request_check_connections(gprs_private *priv, int net_reg_status)
{
  if (is_roaming(priv) && !reg_status_is_roaming(net_reg_status))
  {
    if (gprs_roaming_enabled(priv))
      ILOG_INFO("[GPRS] roaming is allowed, letting connections be open");
    else
    {
      ILOG_INFO("[GPRS] roaming is not allowed, closing down connections");
      close_all_contexts(priv, "Started roaming", NULL);
    }
  }
}

static void
network_request_registration_status_cb(DBusPendingCall *pending,
                                       void *user_data)
{
  gprs_private *priv = user_data;

  enum observer_status status = OBSERVER_STATUS_ERROR;
  guchar prev_status;
  dbus_uint32_t network_country_code;
  dbus_uint32_t network_operator_code;
  dbus_uint32_t network_current_cell_id;
  dbus_int32_t error_value;
  dbus_uint16_t network_current_lac;
  guchar network_supported_services;
  guchar network_type;
  guchar network_reg_status;

  ILOG_INFO("[GPRS] > network_request_registration_status_cb");

  if (pending)
  {
    DBusMessage *reply = dbus_pending_call_steal_reply(pending);

    dbus_pending_call_unref(pending);

    if(reply)
    {
      if (dbus_message_get_type(reply) != DBUS_MESSAGE_TYPE_ERROR)
      {
        if (dbus_message_get_args(reply, NULL,
                                  DBUS_TYPE_BYTE, &network_reg_status,
                                  DBUS_TYPE_UINT16, &network_current_lac,
                                  DBUS_TYPE_UINT32, &network_current_cell_id,
                                  DBUS_TYPE_UINT32, &network_operator_code,
                                  DBUS_TYPE_UINT32, &network_country_code,
                                  DBUS_TYPE_BYTE, &network_type,
                                  DBUS_TYPE_BYTE, &network_supported_services,
                                  DBUS_TYPE_INT32, &error_value,
                                  DBUS_TYPE_INVALID))
        {

          prev_status = priv->net_reg_status;
          priv->net_reg_status = network_reg_status;
          network_request_check_connections(priv, prev_status);
          status = OBSERVER_STATUS_OK;
          ILOG_INFO("[GPRS] new network reg status: 0x%02x\n",
                    priv->net_reg_status);
        }
        else
          ILOG_WARN("[GPRS] could not get args from D-Bus message");
      }
      else
        ILOG_WARN("[GPRS] 'get_registration_status' call returned '%s'",
                  dbus_message_get_error_name(reply));

      dbus_message_unref(reply);
    }
  }

  notify_observers(OBSERVER_TYPE_2, status, priv);
}

gboolean
network_request_registration_status(gprs_private *priv)
{
  DBusMessage *message;
  DBusPendingCall *mcall;

  ILOG_INFO("[GPRS] > network_request_registration_status");

  if (priv->observers[OBSERVER_TYPE_NET_STATUS].active)
    return TRUE;

  message = dbus_message_new_method_call("com.nokia.phone.net",
                                         "/com/nokia/phone/net",
                                         "Phone.Net",
                                         "get_registration_status");
  if (!message)
  {
    ILOG_ERR("[GPRS] could not create 'get_registration_status' method call");
    return FALSE;
  }

  mcall = icd_dbus_send_system_mcall(message, 10000,
                                     network_request_registration_status_cb,
                                     priv);

  if (mcall)
  {
    priv->observers[OBSERVER_TYPE_NET_STATUS].mcall = mcall;
    priv->observers[OBSERVER_TYPE_NET_STATUS].active = TRUE;
    dbus_message_unref(message);
    ILOG_INFO("[GPRS] < network_request_registration_status (success)");
    return TRUE;
  }
  else
  {
    ILOG_ERR("[GPRS] could not send 'get_registration_status' mcall");
        dbus_message_unref(message);
  }

  return FALSE;
}
