#include <icd/support/icd_dbus.h>
#include <icd/support/icd_log.h>

#include "gprs-private.h"
#include "observer.h"

static void
sim_request_imsi_cb(DBusPendingCall *pending, void *user_data)
{
  gprs_private *priv = user_data;
  enum observer_status status = OBSERVER_STATUS_ERROR;

  ILOG_INFO("[GPRS] > sim_request_imsi_cb");

  if (pending)
  {
    DBusMessage *reply = dbus_pending_call_steal_reply(pending);

    dbus_pending_call_unref(pending);

    if (reply)
    {
      if (dbus_message_get_type(reply) == DBUS_MESSAGE_TYPE_ERROR)
      {
          ILOG_WARN("[GPRS] 'get_imsi' call returned '%s'",
                    dbus_message_get_error_name(reply));
      }
      else
      {
        dbus_int32_t error_value;
        gchar *imsi = NULL;

        if (dbus_message_get_args(reply, NULL,
                                  DBUS_TYPE_STRING, &imsi,
                                  DBUS_TYPE_INT32, &error_value,
                                  DBUS_TYPE_INVALID))
        {
          if (error_value)
            ILOG_WARN("[GPRS] IMSI error: %d", error_value);
          else
          {
            priv->imsi = g_strdup(imsi);
            status = OBSERVER_STATUS_OK;
          }
        }
        else
          ILOG_WARN("[GPRS] could not get args from D-Bus message");
      }
    }

    dbus_message_unref(reply);
  }

  notify_observers(OBSERVER_TYPE_SIM_IMSI, status, priv);

  ILOG_INFO("[GPRS] < sim_request_imsi_cb");
}

gboolean
sim_request_imsi(gprs_private *priv)
{
  DBusMessage *mcall;
  DBusPendingCall *pending;

  ILOG_INFO("[GPRS] > sim_request_imsi");

  if (priv->imsi_requested)
    return TRUE;

  mcall = dbus_message_new_method_call("com.nokia.phone.SIM",
                                       "/com/nokia/phone/SIM",
                                       "Phone.Sim",
                                       "get_imsi");

  if (!mcall)
  {
    ILOG_ERR("[GPRS] could not create 'get_imsi' method call");
    return FALSE;
  }

  pending = icd_dbus_send_system_mcall(mcall, 10000, sim_request_imsi_cb, priv);

  if (pending)
  {
    priv->sim_request_imsi_pending = pending;
    priv->imsi_requested = TRUE;
    dbus_message_unref(mcall);
    ILOG_INFO("[GPRS] < sim_request_imsi");
    return TRUE;
  }

  ILOG_ERR("[GPRS] could not send 'get_imsi' mcall");
  dbus_message_unref(mcall);

  return FALSE;
}

static void
sim_request_status_cb(DBusPendingCall *pending, void *user_data)
{
  gprs_private *priv = user_data;
  enum observer_status status = OBSERVER_STATUS_ERROR;

  ILOG_INFO("[GPRS] > sim_request_status_cb");

  if (pending)
  {
    DBusMessage *reply = dbus_pending_call_steal_reply(pending);

    dbus_pending_call_unref(pending);

    if (reply)
    {
      if (dbus_message_get_type(reply) == DBUS_MESSAGE_TYPE_ERROR)
      {
          ILOG_WARN("[GPRS] 'get_sim_status' call returned '%s'",
                    dbus_message_get_error_name(reply));
      }
      else
      {
        dbus_int32_t error_value;
        dbus_uint32_t sim_status;

        if (dbus_message_get_args(reply, NULL,
                                  DBUS_TYPE_UINT32, &sim_status,
                                  DBUS_TYPE_INT32, &error_value,
                                  DBUS_TYPE_INVALID))
        {
          if (!error_value)
          {
            status = OBSERVER_STATUS_OK;
            priv->sim_status = sim_status;
          }
          else
            ILOG_WARN("[GPRS] SIM status query error: %d", error_value);
        }
        else
          ILOG_WARN("[GPRS] could not get args from D-Bus message");
      }

      dbus_message_unref(reply);
    }
  }

  notify_observers(OBSERVER_TYPE_NET_STATUS, status, priv);

  ILOG_INFO("[GPRS] < sim_request_status_cb");
}

gboolean
sim_request_status(gprs_private *priv)
{
  DBusMessage *message;
  DBusPendingCall *mcall;

  if (priv->observers[0].active == TRUE)
    return TRUE;

  message = dbus_message_new_method_call("com.nokia.phone.SIM",
                                         "/com/nokia/phone/SIM",
                                         "Phone.Sim",
                                         "get_sim_status");
  if (message)
  {
    mcall = icd_dbus_send_system_mcall(message, 10000,
                                     sim_request_status_cb, priv);
    if (mcall)
    {
      priv->observers[OBSERVER_TYPE_SIM_IMSI].mcall = mcall;
      priv->observers[OBSERVER_TYPE_SIM_IMSI].active = TRUE;
      dbus_message_unref(message);
      return TRUE;
    }
    else
    {
      ILOG_ERR("[GPRS] could not send 'get_sim_status' mcall");
      dbus_message_unref(message);
    }
  }
  else
    ILOG_ERR("[GPRS] could not create 'get_sim_status' method call");

  return FALSE;
}
